from __future__ import annotations
import materia

__all__ = ["MultiwfnInput"]


class MultiwfnInput:
    """
    Interface for inputting commands to Multiwfn.

    Attributes:
        commands (str): Iterable of commands which will be fed sequentially to Multiwfn.
    """

    def __init__(self, filepath: str, *commands: str) -> None:
        self.commands = tuple((materia.expand(filepath), *commands))

    def write(self, filepath: str) -> None:
        """
        Write Multiwfn input to a file.

        Args:
            filepath: Path to file to which the input will be written. Can be an absolute or a relative path.
        """
        with open(materia.expand(filepath), "w") as f:
            f.write(str(self))

    def __str__(self) -> str:
        return "\n".join(str(c) for c in self.commands) + "\n"



from __future__ import annotations
import os
import materia
import re


__all__ = ["VASPInput"]

# FIXME: incomplete - need to add _potcar_str


class VASPInput:
    def __init__(self, settings: materia.Settings) -> None:
        self.settings = settings

    def write(self, directory: str) -> None:
        with open(materia.expand(os.path.join(directory), "INCAR"), "w") as f:
            f.write(_incar_str(self.settings["INCAR"]))
        with open(materia.expand(os.path.join(directory), "POSCAR"), "w") as f:
            f.write(_poscar_str(self.settings["POSCAR"]))
        with open(materia.expand(os.path.join(directory), "KPOINTS"), "w") as f:
            f.write(_kpoints_str(self.settings["KPOINTS"]))
        with open(materia.expand(os.path.join(directory), "POTCAR"), "w") as f:
            f.write(_potcar_str(self.settings["POTCAR"]))

    def __str__(self) -> str:
        return "\n".join(
            _block_to_str(block_name=block_name, block_params=block_params)
            for block_name, block_params in self.settings.d.items()
        )


def _block_to_str(
    block_name: str, block_params
) -> str:  # FIXME: unsure how to annotate type hint for block_params...
    if block_name.lower() == "incar":
        return _incar_str(block_params)
    elif block_name.lower() == "poscar":
        return _poscar_str(block_params)
    elif block_name.lower() == "kpoints":
        return _kpoints_str(block_params)
    elif block_name.lower() == "potcar":
        return _potcar_str(block_params)
    else:
        return ""


def _incar_str(settings: materia.Settings) -> str:
    s = ""
    try:
        s += str(settings["comment"])
    except KeyError:
        pass
    s += "\n"
    s += (
        "\n".join(
            f"{k} = {v}"
            if (isinstance(v, str) or not hasattr(v, "__iter__"))
            else f"{k} = " + " ".join(f"{x}" for x in v)
            for k, v in settings.items()
        )
        .replace("True", ".TRUE.")
        .replace("False", ".FALSE.")
    )
    return s


def _kpoints_str(settings: materia.Settings) -> str:
    s = ""
    try:
        s += str(settings["comment"])
    except KeyError:
        pass

    s += "\n"

    try:
        s += str(settings["num_kpoints"]) + "\n"
    except KeyError:
        pass

    try:
        s += str(settings["mesh_type"]) + "\n"
    except KeyError:
        pass

    try:
        num_x, num_y, num_z = settings["grid"]
        s += f"{num_x}  {num_y}  {num_z}\n"
    except KeyError:
        pass

    try:
        shift_x, shift_y, shift_z = settings["shift"]
        s += f"{shift_x}  {shift_y}  {shift_z}\n"
    except KeyError:
        pass

    return s


def _poscar_str(settings: materia.Settings) -> str:
    s = ""
    try:
        s += str(settings["comment"])
    except KeyError:
        pass

    s += "\n"

    try:
        s += str(settings["scaling"]) + "\n"
    except KeyError:
        pass

    try:
        a_1, a_2, a_3, b_1, b_2, b_3, c_1, c_2, c_3 = settings["bravais_matrix"]
        s += f"{a_1}  {a_2}  {a_3}\n{b_1}  {b_2}  {b_3}\n{c_1}  {c_2}  {c_3}\n"
    except KeyError:
        pass

    try:
        s += " ".join(str(n) for n in settings["num_atoms"]) + "\n"
    except KeyError:
        pass

    try:
        if settings["direct"]:
            direct = True
            s += "Direct\n"
    except KeyError:
        direct = False
    try:
        if settings["cartesian"]:
            cartesian = True
            s += "Cartesian\n"
    except KeyError:
        cartesian = False

    try:  # FIXME: incorporate direct vs cartesian here (i.e. fractionalize when direct = True)
        structure = settings["structure"]
        s += "\n".join((f"{x}  {y}  {z}" for (x, y, z) in structure.atomic_positions.T))
    except KeyError:
        pass

    return s


# import numpy as np
# import materia
# from materia.workflow.interfaces.vasp.input import _incar_str,_kpoints_str,_poscar_str,VASPInput
# s = materia.Settings()
# s['incar','PREC'] = 'Normal'
# s['incar','ENCUT'] = 550
# s['incar','IBRION'] = -1
# s['incar','NSW'] = 0
# s['incar','ISIF'] = 2
# s['incar','NELMIN'] = 2
# s['incar','EDIFF'] = 1.0e-5
# s['incar','EDIFFG' = -0.02
# s['incar','VOSKOWN'] = 1
# s['incar','NBLOCK'] = 1
# s['incar','NWRITE'] = 1
# s['incar','NELM'] = 60
# s['incar','LUSE_VDW'] = True
# s['incar','LASPH'] = True
# s['incar','AGGAC'] = 0.0000
# s['incar','GGA'] = 'MK'
# s['incar','PARAM1'] = 0.1234
# s['incar','PARAM2'] = 1.0000
# s['incar','ALGO'] = 'Normal (blocked Davidson)'
# s['incar','ISPIN'] = 2
# s['incar','INIWAV'] = 1
# s['incar','ISTART'] = 0
# s['incar','ICHARG'] = 2
# s['incar','LWAVE'] = False
# s['incar','LCHARG'] = False
# s['incar','ADDGRID'] = False
# s['incar','ISMEAR'] = 1
# s['incar','SIGMA'] = 0.2
# s['incar','LREAL'] = False
# s['incar','RWIGS'] = (1.17,0.73)
# s['incar','LDAU'] = True
# s['incar','LDAUTYPE'] = 2
# s['incar','LDAUL'] = (2,-1)
# s['incar','LDAUU'] = (8.00,0.00)
# s['incar','LDAUJ'] = (4.00,0.00)
# s['incar','LDAUPRINT'] = 2
# s['poscar','comment'] = '(Fe2 O3)24  (P1) ~ (COD #9015065)_1_2x2x1_1 0 0 1  surface_1 (#1)_2x2x1_1 (MD #5) (VASP)'
# s['poscar','scaling'] = 1.0
# s['poscar','bravais_matrix'] = np.hstack([[10.0498,0,0],[0,8.7034,0],[0,0,28.7163]])
# s['poscar','Direct'] = True
# s['poscar','num_atoms'] = (48,72)
# s['kpoints','comment'] = 'Automatic grid'
# s['kpoints','mesh_type'] = 'Gamma'
# s['kpoints','grid'] = (4,4,1)
# s['kpoints','shift'] = (0.0,0.0,0.0)
# print(VASPInput(s))



from __future__ import annotations
from typing import Any, Dict, Iterable, Optional, Tuple, Union

import materia

__all__ = [
    "GAMESSBasis",
    "GAMESSContrl",
    "GAMESSDamp",
    "GAMESSDampGS",
    "GAMESSSCF",
    "GAMESSStone",
    "GAMESSStructure",
    "GAMESSSystem",
]


class GAMESSBasis:
    def __init__(
        self,
        gbasis: Optional[str] = None,
        ngauss: Optional[int] = None,
        npfunc: Optional[int] = None,
        ndfunc: Optional[int] = None,
        nffunc: Optional[int] = None,
        diffs: Optional[bool] = None,
        diffsp: Optional[bool] = None,
    ) -> None:
        self.gbasis = gbasis
        self.ngauss = ngauss
        self.npfunc = npfunc
        self.ndfunc = ndfunc
        self.nffunc = nffunc
        self.diffs = diffs
        self.diffsp = diffsp

    def _process(self, k: str, v: Any) -> Tuple[str, Any]:
        if isinstance(v, bool):
            return k, f".{str(v)}."
        return k, v

    def __str__(self) -> str:
        keywords = (self._process(k, v) for k, v in vars(self).items())
        return (
            " $basis\n"
            + " \n".join(f"{k}={v}" for k, v in keywords if v is not None)
            + "\n $end\n"
        )


class GAMESSContrl:
    def __init__(
        self,
        runtyp: Optional[str] = None,
        units: Optional[Union[str, materia.Qty]] = None,
        coord: Optional[str] = None,
        icut: Optional[int] = None,
        local: Optional[str] = None,
    ) -> None:
        self.runtyp = runtyp
        self.units = units
        self.coord = coord
        self.icut = icut
        self.local = local

    def _process(self, k: str, v: Any) -> Tuple[str, Any]:
        if k == "units" and isinstance(v, materia.Qty):
            if v == materia.angstrom:
                return k, "angs"
            elif v == materia.bohr:
                return k, "bohr"
            else:
                return k, None
        return k, v

    def __str__(self) -> str:
        keywords = (self._process(k, v) for k, v in vars(self).items())
        return (
            " $contrl\n"
            + " \n".join(f"{k}={v}" for k, v in keywords if v is not None)
            + "\n $end\n"
        )


class GAMESSDamp:
    def __init__(
        self,
        ifttyp: Optional[Union[int, Iterable[int]]] = None,
        iftfix: Optional[Union[int, Iterable[int]]] = None,
        thrsh: Optional[Union[float, materia.Qty]] = None,
    ) -> None:
        self.ifttyp = ifttyp
        self.iftfix = iftfix
        self.thrsh = thrsh

    def _process(self, k: str, v: Any) -> Tuple[str, Any]:
        if k in ("ifttyp", "iftfix"):
            try:
                return f"{k}(1)", ",".join(str(i) for i in v)
            except TypeError:
                return f"{k}(1)", v

        if k == "thrsh" and isinstance(v, materia.Qty):
            return k, v.convert(materia.kcal / materia.mol).value

        return k, v

    def __str__(self) -> str:
        keywords = (self._process(k, v) for k, v in vars(self).items())
        return (
            " $damp\n"
            + " \n".join(f"{k}={v}" for k, v in keywords if v is not None)
            + "\n $end\n"
        )


class GAMESSDampGS:
    def __init__(
        self,
        equate_symmetric_points: Optional[bool] = False,
        instructions: Optional[Iterable[str]] = None,
    ) -> None:
        self.equate_symmetric_points = equate_symmetric_points
        self.instructions = instructions or []

    def __str__(self) -> str:
        if self.equate_symmetric_points:
            raise NotImplementedError
        else:
            keywords = (self._process(k, v) for k, v in vars(self).items())
            return " $dampgs\n" + " \n".join(self.instructions) + "\n $end\n"


class GAMESSSCF:
    def __init__(
        self,
        soscf: Optional[bool] = None,
        diis: Optional[bool] = None,
        conv: Optional[float] = None,
    ) -> None:
        self.soscf = soscf
        self.diis = diis
        self.conv = conv

    def _process(self, k: str, v: Any) -> Tuple[str, Any]:
        if isinstance(v, bool):
            return k, f".{str(v)}."
        if k == "conv" and v is not None:
            return k, str(v).replace("e", "d")
        return k, v

    def __str__(self) -> str:
        keywords = (self._process(k, v) for k, v in vars(self).items())
        return (
            " $scf\n"
            + " \n".join(f"{k}={v}" for k, v in keywords if v is not None)
            + "\n $end\n"
        )


class GAMESSStone:
    def __init__(self, bigexp: Optional[int] = None) -> None:
        self.bigexp = bigexp

    def _process(self, k: str, v: Any) -> Tuple[str, Any]:
        return k, v

    def __str__(self) -> str:
        keywords = (self._process(k, v) for k, v in vars(self).items())
        return (
            " $stone\n"
            + " \n".join(f"{k}={v}" for k, v in keywords if v is not None)
            + "\n $end\n"
        )


class GAMESSStructure:
    def __init__(self, structure: materia.Structure) -> None:
        self.structure = structure

    def __str__(self) -> str:
        return (
            " $data\n\nc1\n"
            + "\n".join(
                f"{a.atomic_symbol.lower()}{i+1} {float(a.Z)} "
                + "  ".join(str(p) for p in a.position.reshape(3,))
                for i, a in enumerate(self.structure.atoms)
            )
            + "\n $end"
        )


class GAMESSSystem:
    def __init__(
        self,
        mwords: Optional[int] = None,
        timlim: Optional[Union[float, materia.Qty]] = None,
    ) -> None:
        self.mwords = mwords
        self.timlim = timlim

    def _process(self, k: str, v: Any) -> Tuple[str, Any]:
        if k == "timlim" and isinstance(v, materia.Qty):
            return k, v.convert(materia.minute).value

        return k, v

    def __str__(self) -> str:
        keywords = (self._process(k, v) for k, v in vars(self).items())
        return (
            " $system\n"
            + " \n".join(f"{k}={v}" for k, v in keywords if v is not None)
            + "\n $end\n"
        )



from .astmg173 import ASTMG173
from .blackbody_spd import BlackbodySPD
from .cie_cmfs import (
    CIE1931ColorMatchingFunctionX,
    CIE1931ColorMatchingFunctionY,
    CIE1931ColorMatchingFunctionZ,
)
from .cie_illuminants import (
    CIEIlluminantA,
    CIEIlluminantD65,
    CIEIlluminantDSeries,
    CIEIlluminantF4,
)
from .cie_test_color_samples import (
    CIE1995TestColorSample01,
    CIE1995TestColorSample02,
    CIE1995TestColorSample03,
    CIE1995TestColorSample04,
    CIE1995TestColorSample05,
    CIE1995TestColorSample06,
    CIE1995TestColorSample07,
    CIE1995TestColorSample08,
)
from .dataseries import *
from .simmons_d_series import SimmonsDSeriesPCA0, SimmonsDSeriesPCA1, SimmonsDSeriesPCA2



import abc
import functools

import materia as mtr

__all__ = ["Handler"]


class Handler(abc.ABC):
    def run(self, result, task):
        if self.check(result=result, task=task):
            raise mtr.ActionSignal(
                result=result, actions=self.handle(result=result, task=task)
            )

    @abc.abstractmethod
    def check(self, result, task):
        # input: result of task to be checked and the task object itself
        # output: True if output indicates that self.handle should be called, else False
        pass

    @abc.abstractmethod
    def handle(self, result, task):
        # input: result of task to be checked and the task object itself
        # output: Actions object containing actions to be performed in order to repair task
        return []



from __future__ import annotations
import inspect
import materia
from typing import Tuple, TypeVar, Union

__all__ = ["PackmolOutput"]

T = TypeVar("T")


class PackmolOutput:
    def __init__(self, filepath: str) -> None:
        # raise NotImplementedError
        self.filepath = materia.expand(filepath)

    def get(self, *quantity_names: str) -> Union[T, Tuple[T]]:
        method_dict = dict(inspect.getmembers(self, predicate=inspect.isroutine))

        with open(self.filepath, "r") as f:
            lines = "".join(f.readlines())

        quantities = tuple(
            method_dict[name.lower()](lines=lines) for name in quantity_names
        )

        if len(quantities) == 1:
            return quantities[0]
        return quantities



from __future__ import annotations
import materia
from typing import Optional

__all__ = [
    "NWChemInput",
    "NWChemTopBlock",
    "NWChemTopMemorySubblock",
    "NWChemTopSetSubblock",
    "NWChemTopUnsetSubblock",
    "NWChemTopChargeSubblock",
    "NWChemTopTaskSubblock",
    "NWChemGeometryBlock",
    "NWChemBasisBlock",
    "NWChemDFTBlock",
    "NWChemDFTSemidirectSubblock",
    "NWChemDFTConvergenceSubblock",
    "NWChemDFTXCSubblock",
    "NWChemDFTGridSublock",
    "NWChemDFTVectorsSubblock",
    "NWChemRTTDDFTBlock",
    "NWChemRTTDDFTFieldSubblock",
    "NWChemRTTDDFTExciteSubblock",
    "NWChemRTTDDFTVisualizationSubblock",
    "NWChemCosmoBlock",
    "NWChemRamanBlock",
    "NWChemPropertyBlock",
]


class NWChemInput:
    def __init__(self) -> None:
        self.input_strings = []

        self.block_class_lookup_dict = {
            "top": NWChemTopBlock,
            "geometry": NWChemGeometryBlock,
            "basis": NWChemBasisBlock,
            "dft": NWChemDFTBlock,
            "rttddft": NWChemRTTDDFTBlock,
            "cosmo": NWChemCosmoBlock,
            "raman": NWChemRamanBlock,
            "property": NWChemPropertyBlock,
        }

    def __str__(self) -> str:
        return "\n".join(self.input_strings).rstrip()

    def add_block(
        self, **block_params
    ) -> None:  # FIXME: not sure how to annotate **block_params with type hints...
        self.input_strings.append(
            "\n".join(
                str(self.block_class_lookup_dict[k](**v))
                for k, v in block_params.items()
            )
        )

    def write(self, filepath: str) -> None:
        with open(materia.expand(filepath), "w") as f:
            f.write(str(self))


# TOP LEVEL


class NWChemTopBlock:
    # print & noprint not implemented
    def __init__(  # FIXME: not sure how to annotate all of these with type hints...
        self,
        title: Optional[str] = None,
        echo: bool = False,
        scratch_dir: Optional[str] = None,
        permanent_dir: Optional[str] = None,
        start: bool = False,
        restart: bool = False,
        memory_params=None,
        set_params=None,
        unset_params=None,
        task_params=None,
    ):
        self.title = title
        self.echo = echo
        self.scratch_dir = scratch_dir
        self.permanent_dir = permanent_dir
        self.start = start
        self.restart = restart
        self.memory_subblock = (
            NWChemTopMemorySubblock(**memory_params)
            if memory_params is not None
            else None
        )
        self.set_subblock = (
            NWChemTopSetSubblock(**set_params) if set_params is not None else None
        )
        self.unset_subblock = (
            NWChemTopUnsetSubblock(**unset_params) if unset_params is not None else None
        )
        self.task_subblock = (
            NWChemTopTaskSubblock(**task_params) if task_params is not None else None
        )

    def __str__(self) -> str:
        s = []
        s.append(f'title "{self.title}"\n' if self.title is not None else "")
        s.append("echo\n" if self.echo else "")
        s.append(
            f"scratch_dir {self.scratch_dir}\n" if self.scratch_dir is not None else ""
        )
        s.append(
            f"permanent_dir {self.permanent_dir}\n"
            if self.permanent_dir is not None
            else ""
        )
        s.append(str(self.memory_subblock) if self.memory_subblock is not None else "")
        s.append(str(self.set_subblock) if self.set_subblock is not None else "")
        s.append(str(self.unset_subblock) if self.unset_subblock is not None else "")
        s.append(str(self.task_subblock) if self.task_subblock is not None else "")
        s.append(f'start "{self.title}"\n' if self.start else "")
        s.append(f'restart "{self.title}"\n' if self.restart else "")

        return "".join(s)


class NWChemTopMemorySubblock:
    def __init__(
        self,
        total_mem=None,
        total_units=None,
        stack_mem=None,
        stack_units=None,
        global_mem=None,
        global_units=None,
        heap_mem=None,
        heap_units=None,
    ):
        self.total_mem = total_mem
        self.total_units = total_units
        self.stack_mem = stack_mem
        self.stack_units = stack_units
        self.global_mem = global_mem
        self.global_units = global_units
        self.heap_mem = heap_mem
        self.heap_units = heap_units

    def __str__(self) -> str:
        s = []
        if not (
            self.total_mem is None
            and self.stack_mem is None
            and self.global_mem is None
            and self.heap_mem is None
        ):
            s.append("memory")
        if self.total_mem is not None and self.total_units is not None:
            s.append(f" total {self.total_mem} {self.total_units}")
        if self.stack_mem is not None and self.stack_units is not None:
            s.append(f" stack {self.stack_mem} {self.stack_units}")
        if self.global_mem is not None and self.global_units is not None:
            s.append(f" global {self.global_mem} {self.global_units}")
        if self.heap_mem is not None and self.heap_units is not None:
            s.append(f" heap {self.heap_mem} {self.heap_units}")
        s.append("\n")

        return "".join(s)


class NWChemTopSetSubblock:
    def __init__(self, variable_name, variable_value, variable_type=None) -> None:
        self.variable_name = variable_name
        self.variable_value = variable_value
        self.variable_type = variable_type

    def __str__(self) -> str:
        if self.variable_type is not None:
            return (
                f"set {self.variable_name} {self.variable_type} {self.variable_value}\n"
            )
        else:
            return f"set {self.variable_name} {self.variable_value}\n"


class NWChemTopUnsetSubblock:
    def __init__(self, variable_name) -> None:
        self.variable_name = variable_name

    def __str__(self) -> str:
        return f"unset {self.variable_name}\n"


class NWChemTopChargeSubblock:
    def __init__(self, charge) -> None:
        self.charge = charge

    def __str__(self) -> str:
        return f"charge {self.charge}\n"


class NWChemTopTaskSubblock:
    def __init__(self, theory_level, operation=None, ignore=False) -> None:
        self.theory_level = theory_level
        self.operation = operation
        self.ignore = ignore

    def __str__(self) -> str:
        s = []
        s.append(f"task {self.theory_level}")
        s.append(f" {self.operation}" if self.operation is not None else "")
        s.append(" ignore" if self.ignore else "")
        s.append("\n")

        return "".join(s)


# GEOMETRY


class NWChemGeometryBlock:
    def __init__(
        self,
        title,
        xyz_filepath,
        units="angstroms",
        center=True,
        autosym=True,
        autoz=True,
    ):
        self.title = title
        self.xyz_filepath = xyz_filepath
        self.units = units
        self.center = "center" if center else "nocenter"
        self.autosym = "autosym" if autosym else "noautosym"
        self.autoz = "autoz" if autoz else "noautoz"

    def __str__(self) -> str:
        s = []
        s.append(
            f'geometry "{self.title}" units {self.units} {self.center} {self.autosym} {self.autoz}\n'
        )
        s.append(f'  LOAD "{self.xyz_filepath}"\n')
        s.append("end\n")

        return "".join(s)


# BASIS


class NWChemBasisBlock:
    def __init__(
        self,
        atom_strings,
        library_strings,
        angular_type="cartesian",
        segment=True,
        print_bool=True,
    ):
        self.atom_strings = atom_strings
        self.library_strings = library_strings
        self.angular_type = angular_type
        self.segment = "segment" if segment else "nosegment"
        self.print_bool = "print" if print_bool else "noprint"

    def __str__(self) -> str:
        s = []
        s.append(f"basis {self.angular_type} {self.segment} {self.print_bool}\n")
        for atom, library in zip(self.atom_strings, self.library_strings):
            s.append(f"  {atom} library {library}\n")
        s.append("end\n")

        return "".join(s)


# DFT


class NWChemDFTBlock:
    # cdft,fon,occup,tolerances,cgmin,rodft,sic,mulliken,fukui,bsse,disp,noscf,print,sodft,max_ovel not implemented yet
    def __init__(
        self,
        iterations=None,
        smear=None,
        direct=None,
        incore=None,
        noio=None,
        odft=False,
        mult=None,
        semidirect_params=None,
        xc_params=None,
        convergence_params=None,
        grid_params=None,
        vectors_params=None,
    ):
        self.iterations = iterations
        self.smear = smear
        self.direct = direct
        self.incore = incore
        self.noio = noio
        self.odft = odft
        self.mult = mult

        self.semidirect_subblock = (
            NWChemDFTSemidirectSubblocbk(**semidirect_params)
            if semidirect_params is not None
            else None
        )
        self.xc_subblock = (
            NWChemDFTXCSubblock(**xc_params) if xc_params is not None else None
        )
        self.convergence_subblock = (
            NWChemDFTConvergenceSubblock(**convergence_params)
            if convergence_params is not None
            else None
        )
        self.grid_subblock = (
            NWChemDFTGridSubblock(**grid_params) if grid_params is not None else None
        )
        self.vectors_subblock = (
            NWChemDFTVectorsSubblock(**vectors_params)
            if vectors_params is not None
            else None
        )

    def __str__(self) -> str:
        s = []
        s.append("dft\n")
        s.append(
            f"  iterations {self.iterations}\n" if self.iterations is not None else ""
        )
        s.append(f"  smear {self.smear}\n" if self.smear is not None else "")
        s.append("  direct\n" if self.direct else "")
        s.append("  incore\n" if self.incore else "")
        s.append("  noio\n" if self.noio else "")
        s.append(
            str(self.semidirect_subblock)
            if self.semidirect_subblock is not None
            else ""
        )
        s.append("  odft\n" if self.odft else "")
        s.append(f"  mult {self.mult}\n" if self.mult is not None else "")
        s.append(str(self.xc_subblock) if self.xc_subblock is not None else "")
        s.append(
            str(self.convergence_subblock)
            if self.convergence_subblock is not None
            else ""
        )
        s.append(str(self.grid_subblock) if self.grid_subblock is not None else "")
        s.append(
            str(self.vectors_subblock) if self.vectors_subblock is not None else ""
        )
        s.append("end\n")

        return "".join(s)


class NWChemDFTSemidirectSubblock:
    def __init__(self, filesize=None, memsize=None, filename=None) -> None:
        self.filesize = filesize
        self.memsize = memsize
        self.filename = filename

    def __str__(self) -> str:
        s = []
        s.append("  semidirect")
        s.append(f" filesize {self.filesize}" if self.filesize is not None else "")
        s.append(f" memsize {self.memsize}" if self.memsize is not None else "")
        s.append(f" filename {self.filename}" if self.filename is not None else "")
        s.append("\n")

        return "".join(s)


class NWChemDFTConvergenceSubblock:
    def __init__(
        self,
        energy=None,
        density=None,
        gradient=None,
        hl_tol=None,
        damp=None,
        diis=None,
        lshift=None,
        dampon=None,
        dampoff=None,
        diison=None,
        diisoff=None,
        levlon=None,
        levloff=None,
        ncydp=None,
        ncyds=None,
        ncysh=None,
        rabuck=None,
        nodamping=False,
        nodiis=False,
        nolevelshifting=False,
    ):
        self.energy = energy
        self.density = density
        self.gradient = gradient
        self.hl_tol = hl_tol
        self.damp = damp
        self.diis = diis
        self.lshift = lshift
        self.dampon = dampon
        self.dampoff = dampoff
        self.diison = diison
        self.diisoff = diisoff
        self.levlon = levlon
        self.levloff = levloff
        self.ncydp = ncydp
        self.ncyds = ncyds
        self.ncysh = ncysh
        self.rabuck = rabuck
        self.nodamping = nodamping
        self.nodiis = nodiis
        self.nolevelshifting = nolevelshifting

    def __str__(self) -> str:
        s = []
        s.append("  convergence ")
        if self.energy is not None:
            s.append(f"energy {self.energy} ")
        if self.density is not None:
            s.append(f"density {self.density} ")
        if self.gradient is not None:
            s.append(f"gradient {self.gradient} ")
        if self.hl_tol is not None:
            s.append(f"hl_tol {self.hl_tol} ")
        if self.damp is not None:
            s.append(f"damp {self.damp} ")
        if self.diis is not None:
            s.append(f"diis {self.diis} ")
        if self.lshift is not None:
            s.append(f"lshift {self.lshift} ")
        if self.dampon is not None:
            s.append(f"dampon {self.dampon} ")
        if self.dampoff is not None:
            s.append(f"dampoff {self.dampoff} ")
        if self.diison is not None:
            s.append(f"diison {self.diison} ")
        if self.diisoff is not None:
            s.append(f"diisoff {self.diisoff} ")
        if self.levlon is not None:
            s.append(f"levlon {self.levlon} ")
        if self.levloff is not None:
            s.append(f"levloff {self.levloff} ")
        if self.ncydp is not None:
            s.append(f"ncydp {self.ncydp} ")
        if self.ncyds is not None:
            s.append(f"ncyds {self.ncyds} ")
        if self.ncysh is not None:
            s.append(f"ncysh {self.ncysh} ")
        if self.rabuck is not None:
            s.append(f"rabuck {self.rabuck} ")
        if self.nodamping:
            s.append("nodamping ")
        if self.nodiis:
            s.append("nodiis ")
        if self.nolevelshifting:
            s.append("nolevelshifting ")

        s.append("\n")

        return "".join(s)


class NWChemDFTXCSubblock:
    def __init__(
        self,
        xc_functional_strings=(),
        xc_weights=(),
        xc_local_weights=(),
        xc_nonlocal_weights=(),
        cam=None,
        cam_alpha=None,
        cam_beta=None,
    ) -> None:
        self.xc_functional_strings = xc_functional_strings
        self.xc_weights = xc_weights + (None,) * (
            len(xc_functional_strings) - len(xc_weights)
        )
        self.xc_local_weights = xc_local_weights + (None,) * (
            len(xc_functional_strings) - len(xc_local_weights)
        )
        self.xc_nonlocal_weights = xc_nonlocal_weights + (None,) * (
            len(xc_functional_strings) - len(xc_nonlocal_weights)
        )
        self.cam = cam
        self.cam_alpha = cam_alpha
        self.cam_beta = cam_beta

    def __str__(self) -> str:
        s = []
        s.append("  xc")
        for xc, weight, local_weight, nonlocal_weight in zip(
            self.xc_functional_strings,
            self.xc_weights,
            self.xc_local_weights,
            self.xc_nonlocal_weights,
        ):
            s.append(f" {xc}")
            if weight is not None:
                s.append(f" {weight}")
            if local_weight is not None:
                s.append(f" local {local_weight}")
            if nonlocal_weight is not None:
                s.append(f" nonlocal {nonlocal_weight}")
        s.append("\n")
        if None not in (self.cam, self.cam_alpha, self.cam_beta):
            s.append(
                f"  cam {self.cam} cam_alpha {self.cam_alpha} cam_beta {self.cam_beta}\n"
            )

        return "".join(s)


class NWChemDFTGridSublock:
    def __init__(
        self,
        size=None,
        radial_type=None,
        nodisk=False,
        gausleg_points=(),
        gausleg_atoms=(),
        gausleg_atom_points=(),
        lebedev_points=(),
        lebedev_atoms=(),
        lebedev_atom_points=(),
        partition_function=None,
    ) -> None:

        self.size = size
        self.radial_type = radial_type
        self.nodisk = nodisk
        self.gausleg_points = gausleg_points
        self.gausleg_atoms = gausleg_atoms
        self.gausleg_atom_points = gausleg_atom_points
        self.lebedev_points = lebedev_points
        self.lebedev_atoms = lebedev_atoms
        self.lebedev_atom_points = lebedev_atom_points
        self.partition_function = partition_function

    def __str__(self) -> str:
        s = []
        s.append("  grid")
        s.append(f" {self.size}" if size is not None else "")
        s.append(f" {self.radial_type}")
        s.append(
            " gausleg"
            + " ".join(str(p) for p in self.gausleg_points)
            + " ".join(
                f"{a} {rad} {ang}"
                for a, (rad, ang) in zip(self.gausleg_atoms, self.gausleg_atom_points)
            )
        )
        s.append(
            " lebedev"
            + " ".join(str(p) for p in self.lebedev_points)
            + " ".join(
                f"{a} {rad} {ang}"
                for a, (rad, ang) in zip(self.lebedev_atoms, self.lebedev_atom_points)
            )
        )
        s.append(
            f" {self.partitioning_function}"
            if self.partitioning_function is not None
            else ""
        )
        s.append(" nodisk" if self.nodisk else "")
        s.append("\n")

        return "".join(s)


class NWChemDFTVectorsSubblock:
    def __init__(
        self,
        input_filename=None,
        project_basisname=None,
        project_filename=None,
        fragment_filenames=(),
        reorder_pairs=(),
        swap_pairs=(),
        swap_alpha_pairs=(),
        swap_beta_pairs=(),
        rotate_input_geometry=None,
        rotate_input_movecs=None,
        output_filename=None,
    ) -> None:

        self.input_filename = input_filename
        self.project_basisname = project_basisname
        self.project_filename = project_filename
        self.fragment_filenames = fragment_filenames
        self.reorder_pairs = reorder_pairs
        self.swap_pairs = swap_pairs
        self.swap_alpha_pairs = swap_alpha_pairs
        self.swap_beta_pairs = swap_beta_pairs
        self.rotate_input_geometry = rotate_input_geometry
        self.rotate_input_movecs = rotate_input_movecs
        self.output_filename = output_filename

    def __str__(self) -> str:
        s = []
        s.append("  vectors")
        s.append(
            f" input {self.input_filename}" if self.input_filename is not None else ""
        )
        s.append(
            f" project {self.project_basisname} {self.project_filename}"
            if self.project_basisname is not None and self.project_filename is not None
            else ""
        )
        s.append(
            " fragment"
            + "".join(f" {filename}" for filename in self.fragment_filenames)
            if self.fragment_filenames != ()
            else ""
        )
        s.append(
            " reorder"
            + "".join(
                f" {coeff_num1} {coeff_num2}"
                for (coeff_num1, coeff_num2) in self.reorder_pairs
            )
            if self.reorder_pairs != ()
            else ""
        )
        s.append(
            " swap"
            + "".join(
                f" {swap_num1} {swap_num2})"
                for (swap_num1, swap_num2) in self.swap_pairs
            )
            if self.swap_pairs != ()
            else ""
        )
        s.append(
            " swap alpha"
            + "".join(
                f" {swap_num1} {swap_num2})"
                for (swap_num1, swap_num2) in self.swap_alpha_pairs
            )
            if self.swap_alpha_pairs != ()
            else ""
        )
        s.append(
            " swap beta"
            + "".join(
                f" {swap_num1} {swap_num2})"
                for (swap_num1, swap_num2) in self.swap_beta_pairs
            )
            if self.swap_beta_pairs != ()
            else ""
        )
        s.append(
            f" rotate {self.rotate_input_geometry} {self.rotate_input_movecs}"
            if self.rotate_input_geometry is not None
            and self.rotate_input_movecs is not None
            else ""
        )
        s.append(
            f" output {self.output_filename}"
            if self.output_filename is not None
            else ""
        )
        s.append("\n")

        return "".join(s)


# RTTDDFT


class NWChemRTTDDFTBlock:
    # load_type options: 'scf','vectors'
    # propagator options: 'euler','rk4','magnus'
    def __init__(
        self,
        tmax=None,
        dt=None,
        tag=None,
        load_type=None,
        load_from=None,
        nchecks=None,
        nprints=None,
        nrestarts=None,
        tolerances_zero=None,
        tolerances_series=None,
        tolerances_interpol=None,
        propagator=None,
        exp=None,
        prof=False,
        noprop=False,
        static=False,
        print_dipole=False,
        print_quadrupole=False,
        print_field=False,
        print_moocc=False,
        print_energy=False,
        print_cputime=False,
        print_charge=False,
        print_convergence=False,
        print_s2=False,
        print_dipcontribs=False,
        print_all=False,
        field_params=None,
        excite_params=None,
        visualization_params=None,
    ) -> None:

        self.tmax = tmax
        self.dt = dt
        self.tag = tag
        self.load_type = load_type
        self.load_from = load_from
        self.nchecks = nchecks
        self.nprints = nprints
        self.nrestarts = nrestarts
        self.tolerances_zero = tolerances_zero
        self.tolerances_series = tolerances_series
        self.tolerances_interpol = tolerances_interpol
        self.propagator = propagator
        self.exp = exp
        self.prof = prof
        self.noprop = noprop
        self.static = static
        self.print_dipole = print_dipole
        self.print_quadrupole = print_quadrupole
        self.print_field = print_field
        self.print_moocc = print_moocc
        self.print_energy = print_energy
        self.print_cputime = print_cputime
        self.print_charge = print_charge
        self.print_convergence = print_convergence
        self.print_s2 = print_s2
        self.print_dipcontribs = print_dipcontribs
        self.print_all = print_all
        self.field_subblock = (
            NWChemRTTDDFTFieldSubblock(**field_params)
            if field_params is not None
            else None
        )
        self.excite_subblock = (
            NWChemRTTDDFTExciteSubblock(**excite_params)
            if excite_params is not None
            else None
        )
        self.visualization_subblock = (
            NWChemRTTDDFTVisualizationSubblock(**visualization_params)
            if visualization_params is not None
            else None
        )

    def __str__(self) -> str:
        s = []
        s.append("rttddft\n")
        s.append(f"  tmax {self.tmax}\n" if self.tmax is not None else "")
        s.append(f"  dt {self.dt}\n" if self.dt is not None else "")
        s.append(f"  tag {self.tag}\n" if self.tag is not None else "")
        s.append(
            f"  load {self.load_type} {self.load_from}\n"
            if self.load_type is not None and self.load_from is not None
            else ""
        )
        s.append(f"  nchecks {self.nchecks}\n" if self.nchecks is not None else "")
        s.append(f"  nprints {self.nprints}\n" if self.nprints is not None else "")
        s.append(
            f"  nrestarts {self.nrestarts}\n" if self.nrestarts is not None else ""
        )
        if not (
            self.tolerances_zero is None
            and self.tolerances_series is None
            and self.tolerances_interpol is None
        ):
            s.append("  tolerances")
            s.append(
                f" zero {self.tolerances_zero}"
                if self.tolerances_zero is not None
                else ""
            )
            s.append(
                f" series {self.tolerances_series}"
                if self.tolerances_series is not None
                else ""
            )
            s.append(
                f" interpol {self.tolerances_interpol}"
                if self.tolerances_interpol is not None
                else ""
            )
            s.append("\n")
        s.append(
            f"  propagator {self.propagator}\n" if self.propagator is not None else ""
        )
        s.append(f"  exp {self.exp}\n" if self.exp is not None else "")
        s.append("  prof\n" if self.prof else "")
        s.append("  noprop\n" if self.noprop else "")
        s.append("  static\n" if self.static else "")
        if (
            self.print_dipole
            or self.print_quadrupole
            or self.print_field
            or self.print_moocc
            or self.print_energy
            or self.print_cputime
            or self.print_charge
            or self.print_convergence
            or self.print_s2
            or self.print_dipcontribs
            or self.print_all
        ):
            s.append("  print")
            s.append(" dipole" if self.print_dipole else "")
            s.append(" quadrupole" if self.print_quadrupole else "")
            s.append(" field" if self.print_field else "")
            s.append(" moocc" if self.print_moocc else "")
            s.append(" energy" if self.print_energy else "")
            s.append(" cputime" if self.print_cputime else "")
            s.append(" charge" if self.print_charge else "")
            s.append(" convergence" if self.print_convergence else "")
            s.append(" s2" if self.print_s2 else "")
            s.append(" dipcontribs" if self.print_dipcontribs else "")
            s.append(" *" if self.print_all else "")
            s.append("\n")
        if self.field_subblock is not None:
            s.append(str(self.field_subblock))
        if self.excite_subblock is not None:
            s.append(str(self.excite_subblock))
        if self.visualization_subblock is not None:
            s.append(str(self.visualization_subblock))
        s.append("end\n")

        return "".join(s)


class NWChemRTTDDFTFieldSubblock:
    def __init__(
        self,
        field_name=None,
        shape=None,
        polarization=None,
        frequency=None,
        center=None,
        width=None,
        max=None,
        spin=None,
    ):

        self.field_name = field_name
        self.shape = shape
        self.polarization = polarization
        self.frequency = frequency
        self.center = center
        self.width = width
        self.max = max
        self.spin = spin

    def __str__(self) -> str:
        s = []
        s.append(f'  field "{self.field_name}"\n')
        s.append(f"    type {self.shape}\n")
        s.append(f"    polarization {self.polarization}\n")
        s.append(
            f"    frequency {self.frequency}\n" if self.frequency is not None else ""
        )
        s.append(f"    center {self.center}\n" if self.center is not None else "")
        s.append(f"    width {self.width}\n" if self.width is not None else "")
        s.append(f"    max {self.max}\n" if self.max is not None else "")
        s.append(f"    spin {self.spin}\n" if self.spin is not None else "")
        s.append("  end\n")

        return "".join(s)


class NWChemRTTDDFTExciteSubblock:
    def __init__(self, geometry_name=None, field_name=None) -> None:
        self.geometry_name = geometry_name
        self.field_name = field_name

    def __str__(self) -> str:
        return f'  excite "{self.geometry_name}" with "{self.field_name}"\n'


class NWChemRTTDDFTVisualizationSubblock:
    def __init__(self, tstart=None, tend=None, treference=None, dplot=False) -> None:
        self.tstart = tstart
        self.tend = tend
        self.treference = treference
        self.dplot = dplot

    def __str__(self) -> str:
        s = []
        s.append("  visualization\n")
        s.append(f"    tstart {self.tstart}\n" if self.tstart is not None else "")
        s.append(f"    tend {self.tend}\n" if self.tend is not None else "")
        s.append(
            f"    treference {self.treference}\n" if self.treference is not None else ""
        )
        s.append(f"    dplot\n" if self.dplot else "")
        s.append("  end\n")

        return "".join(s)


# COSMO


class NWChemCosmoBlock:
    def __init__(
        self,
        dielec=None,
        parameters_file=None,
        radii=(),
        iscren=None,
        minbem=None,
        ificos=None,
        lineq=None,
        zeta=None,
        gamma_s=None,
        sw_tol=None,
        do_gasphase=None,
        do_cosmo_ks=None,
        do_cosmo_yk=None,
    ):

        self.dielec = dielec
        self.parameters_file = parameters_file
        self.radii = radii
        self.iscren = iscren
        self.minbem = minbem
        self.ificos = ificos
        self.lineq = lineq
        self.zeta = zeta
        self.gamma_s = gamma_s
        self.sw_tol = sw_tol
        self.do_gasphase = do_gasphase
        self.do_cosmo_ks = do_cosmo_ks
        self.do_cosmo_yk = do_cosmo_yk

    def __str__(self) -> str:
        s = []
        s.append("cosmo\n")
        s.append(f"  dielec {self.dielec}\n" if self.dielec is not None else "")
        s.append(
            f"  parameters {self.parameters_file}\n"
            if self.parameters_file is not None
            else ""
        )
        if self.radii != ():
            s.append(f"  radius {self.radii[0]}\n")
            for radius in self.radii[1:]:
                s.append(f"         {radius}\n")
        s.append(f"  iscren {self.iscren}\n" if self.iscren is not None else "")
        s.append(f"  minbem {self.minbem}\n" if self.minbem is not None else "")
        s.append(f"  ificos {self.ificos}\n" if self.ificos is not None else "")
        s.append(f"  lineq {self.lineq}\n" if self.lineq is not None else "")
        s.append(f"  zeta {self.zeta}\n" if self.zeta is not None else "")
        s.append(f"  gamma_s {self.gamma_s}\n" if self.gamma_s is not None else "")
        s.append(f"  sw_tol {self.sw_tol}\n" if self.sw_tol is not None else "")
        s.append(
            f"  do_gasphase {self.do_gasphase}\n"
            if self.do_gasphase is not None
            else ""
        )
        s.append(
            f"  do_cosmo_ks {self.do_cosmo_ks}\n"
            if self.do_cosmo_ks is not None
            else ""
        )
        s.append(
            f"  do_cosmo_yk {self.do_cosmo_yk}\n"
            if self.do_cosmo_yk is not None
            else ""
        )
        s.append("end\n")

        return "".join(s)


# RAMAN


class NWChemRamanBlock:
    def __init__(
        self,
        normal=False,
        resonance=False,
        lorentzian=False,
        gaussian=False,
        low=None,
        high=None,
        first=None,
        last=None,
        width=None,
        dq=None,
    ):
        self.normal = normal
        self.resonance = resonance
        self.lorentzian = lorentzian
        self.gaussian = gaussian
        self.low = low
        self.high = high
        self.first = first
        self.last = last
        self.width = width
        self.dq = dq

    def __str__(self) -> str:
        s = []
        s.append("raman\n")
        s.append("  normal\n" if self.normal else "")
        s.append("  resonance\n" if self.resonance else "")
        s.append("  lorentzian\n" if self.lorentzian else "")
        s.append("  gaussian\n" if self.gaussian else "")
        s.append(f"  low {self.low}\n" if self.low is not None else "")
        s.append(f"  high {self.high}\n" if self.high is not None else "")
        s.append(f"  first {self.first}\n" if self.first is not None else "")
        s.append(f"  last {self.last}\n" if self.last is not None else "")
        s.append(f"  width {self.width}\n" if self.width is not None else "")
        s.append(f"  dq {self.dq}\n" if self.dq is not None else "")
        s.append("end\n")

        return "".join(s)


# PROPERTY


class NWChemPropertyBlock:
    # gauge options: 'dipole', 'velocity'
    # center options: 'com', 'coc', 'origin', 'arb'
    # molden_norm options: 'janpa', nwchem', 'none'
    # grid not implemented yet
    def __init__(
        self,
        nbofile=False,
        dipole=False,
        quadrupole=False,
        octupole=False,
        mulliken=False,
        esp=False,
        efield=False,
        efieldgrad=False,
        efieldgradz4=False,
        gshift=False,
        aimfile=False,
        moldenfile=False,
        molden_norm=None,
        electrondensity=False,
        vectors_filename=None,
        hyperfine_num_atoms=None,
        hyperfine_atoms=(),
        shielding_num_atoms=None,
        shielding_atoms=(),
        spinspin_num_pairs=None,
        spinspin_atom_pairs=(),
        response_order=None,
        response_frequency=None,
        response_gauge=None,
        response_damping=None,
        response_convergence=None,
        response_orbeta=False,
        response_giao=False,
        response_bdtensor=False,
        response_analysis=False,
        all=False,
        center=None,
        center_coords=None,
    ):

        self.nbofile = nbofile
        self.dipole = dipole
        self.quadrupole = quadrupole
        self.octupole = octupole
        self.mulliken = mulliken
        self.esp = esp
        self.efield = efield
        self.efieldgrad = efieldgrad
        self.efieldgradz4 = efieldgradz4
        self.gshift = gshift
        self.aimfile = aimfile
        self.moldenfile = moldenfile
        self.molden_norm = molden_norm
        self.electrondensity = electrondensity
        self.vectors_filename = vectors_filename
        self.hyperfine_num_atoms = hyperfine_num_atoms
        self.hyperfine_atoms = hyperfine_atoms
        self.shielding_num_atoms = shielding_num_atoms
        self.shielding_atoms = shielding_atoms
        self.spinspin_num_pairs = spinspin_num_pairs
        self.spinspin_atom_pairs = spinspin_atom_pairs
        self.response_order = response_order
        self.response_frequency = response_frequency
        self.response_gauge = response_gauge
        self.response_damping = response_damping
        self.response_convergence = response_convergence
        self.response_orbeta = response_orbeta
        self.response_giao = response_giao
        self.response_bdtensor = response_bdtensor
        self.response_analysis = response_analysis
        self.all = all
        self.center = center
        self.center_coords = center_coords

    def __str__(self) -> str:
        s = []
        s.append("property\n")
        s.append(f"  nbofile\n" if self.nbofile else "")
        s.append(f"  dipole\n" if self.dipole else "")
        s.append(f"  quadrupole\n" if self.quadrupole else "")
        s.append(f"  octupole\n" if self.octupole else "")
        s.append(f"  mulliken\n" if self.mulliken else "")
        s.append(f"  esp\n" if self.esp else "")
        s.append(f"  efield\n" if self.efield else "")
        s.append(f"  efieldgrad\n" if self.efieldgrad else "")
        s.append(f"  efieldgradz4\n" if self.efieldgradz4 else "")
        s.append(f"  gshift\n" if self.gshift else "")
        s.append(f"  electrondensity\n" if self.electrondensity else "")
        if self.hyperfine_num_atoms is not None and self.hyperfine_atoms != ():
            s.append(
                f"  hyperfine {self.hyperfine_num_atoms} "
                + " ".join(f"{a}" for a in self.hyperfine_atoms)
                + "\n"
            )
        if self.shielding_num_atoms is not None and self.shielding_atoms != ():
            s.append(
                f"  shielding {self.shielding_num_atoms} "
                + " ".join(f"{a}" for a in self.shielding_atoms)
                + "\n"
            )
        if self.spinspin_num_pairs is not None and self.spinspin_atom_pairs != ():
            s.append(
                f"  spinspin {self.spinspin_num_pairs} "
                + " ".join(f"{a1} {a2}" for (a1, a2) in self.spinspin_atom_pairs)
                + "\n"
            )
        s.append(
            f"  response {self.response_order} {self.response_frequency}\n"
            if self.response_order is not None and self.response_frequency is not None
            else ""
        )
        s.append(
            f"  {self.response_gauge}\n" if self.response_gauge is not None else ""
        )
        s.append(
            f"  damping {self.response_damping}\n"
            if self.response_damping is not None
            else ""
        )
        s.append(
            f"  convergence {self.response_convergence}\n"
            if self.response_convergence is not None
            else ""
        )
        s.append("  orbeta\n" if self.response_orbeta else "")
        s.append("  giao\n" if self.response_giao else "")
        s.append("  bdtensor\n" if self.response_bdtensor else "")
        s.append("  analysis\n" if self.response_analysis else "")
        s.append(
            f"  vectors {self.vectors_filename}\n"
            if self.vectors_filename is not None
            else ""
        )
        if self.center == "arb":
            s.append(f"  center " + " ".join(f"{c}" for c in self.center_coords) + "\n")
        elif self.center is not None:
            s.append(f"  center {self.center}\n")
        s.append(f"  aimfile {self.aimfile}\n" if self.aimfile else "")
        s.append(f"  moldenfile {self.moldenfile}\n" if self.moldenfile else "")
        s.append(
            f"  molden_norm {self.molden_norm}\n"
            if self.molden_norm is not None
            else ""
        )
        s.append(f"  all\n" if self.all else "")
        s.append("end\n")

        return "".join(s)



from .group_theory import *
from .symmetric_axes_generator import generate_axes
from .symmetry_finder import (
    GraphSymmetryFinder,
    SpglibSymmetryFinder,
)  # ,ExplicitSymmetryFinder
from .symmetry_operation import (
    SymmetryOperation,
    Inversion,
    Reflection,
    ProperRotation,
    ImproperRotation,
)

# from .symmetry import SymmetryOperation,Inversion,Reflection,ProperRotation,ImproperRotation



from __future__ import annotations
from typing import Dict, Iterable, Optional, Tuple, Union

import materia

__all__ = ["QChemFragments", "QChemInput", "QChemStructure"]


class QChemStructure:
    def __init__(
        self,
        structure: materia.Structure,
        charge: Optional[int] = 0,
        multiplicity: Optional[int] = 1,
        mm_parameters: Optional[Dict[materia.Atom, str]] = None,
        bonds: Optional[Dict[int, int]] = None,
    ) -> None:
        self.structure = structure
        self.charge = charge
        self.multiplicity = multiplicity
        self.mm_parameters = mm_parameters or {}
        self.bonds = bonds or {}

    def __str__(self) -> str:
        return f"  {self.charge} {self.multiplicity}\n" + "\n".join(
            (
                f"  {a.atomic_symbol}  "
                + "  ".join(str(p) for p in a.position.reshape(3,))
                + (f" {self.mm_parameters[a]}" if a in self.mm_parameters else "")
                + (
                    " "
                    + "  ".join(
                        str(j + 1)
                        for j in self.bonds[i] + [-1] * (4 - len(self.bonds[i]))
                    )
                    if i in self.bonds
                    else ""
                )
                for i, a in enumerate(self.structure.atoms)
            )
        )


class QChemFragments:
    def __init__(
        self,
        structures: Iterable[materia.Structure],
        charges: Optional[Iterable[int]] = None,
        multiplicities: Optional[Iterable[int]] = None,
        total_charge: Optional[int] = 0,
        total_multiplicity: Optional[int] = 1,
    ) -> None:
        self.structures = structures
        self.charges = charges or (0,) * len(self.structures)
        self.multiplicities = multiplicities or (1,) * len(self.structures)
        self.total_charge = total_charge
        self.total_multiplicity = total_multiplicity

    def __str__(self) -> str:
        return f"  {self.total_charge} {self.total_multiplicity}\n--\n" + "\n--\n".join(
            str(QChemStructure(structure=s, charge=c, multiplicity=m))
            for c, m, s in zip(self.charges, self.multiplicities, self.structures)
        )
        # return (
        #     f"  {self.total_charge} {self.total_multiplicity}\n--\n"
        #     + "--\n".join(
        #         f"  {c} {m}\n"
        #         + "\n".join(
        #             f"  {a.atomic_symbol}  "
        #             + " ".join(str(p) for p in a.position.reshape(3,))
        #             for a in structure.atoms
        #         )
        #         + "\n"
        #         for c, m, structure in zip(
        #             self.charges, self.multiplicities, self.structures
        #         )
        #     )
        # )


class QChemInput:
    def __init__(
        self,
        molecule: Union[
            materia.QChemStructure, materia.QChemFragments, materia.Structure
        ],
        settings: materia.Settings,
    ) -> None:
        self.molecule = molecule
        self.settings = settings

    def write(self, filepath: str) -> None:
        with open(materia.expand(filepath), "w") as f:
            f.write(str(self))

    def __str__(self) -> str:
        molecule_block = (
            str(QChemStructure(self.molecule))
            if isinstance(self.molecule, materia.Structure)
            else str(self.molecule)
        )
        return f"$molecule\n{molecule_block}\n$end\n" + "\n".join(
            _block_to_str(block_name=block_name, block_params=block_params)
            for block_name, block_params in self.settings.d.items()
        )


def _block_to_str(block_name: str, block_params) -> str:
    if block_name == "xc_functional":
        return _xc_functional_str(*block_params)
    else:
        return _block_str(block_name=block_name, **block_params)


def _block_str(
    block_name, **block_params
) -> str:  # FIXME: not sure exactly how to annotate type hint for **block_params
    longest_key_length = max(len(k) for k in block_params)
    return (
        f"${block_name}\n"
        + "".join(
            f"  {k}" + " " * (longest_key_length - len(k) + 1) + f"{v}\n"
            for k, v in block_params.items()
        )
        + "$end\n"
    )


def _xc_functional_str(*functional_tuples: Tuple[str, str, str]) -> str:
    component_types, component_names, component_coefficients = zip(*functional_tuples)
    longest_component_name = max(len(cn) for cn in component_names)

    s = f"$xc_functional\n"
    for ct, cn, cc in zip(component_types, component_names, component_coefficients):
        if ct == "K":
            s += "  K " + " " * (longest_component_name + 1) + f" {cc}\n"
        else:
            s += (
                f"  {ct} {cn}"
                + " " * (longest_component_name - len(cn) + 1)
                + f" {cc}\n"
            )
    s += "$end\n"

    return s



from __future__ import annotations
import collections
import materia
from typing import Iterable, List, Tuple

__all__ = ["NWChemOutput"]


class NWChemOutput:
    def __init__(self, filename: str) -> None:
        self.filename = materia.expand(filename)
        self.output_dict = collections.OrderedDict()

        self.chunk_dict = {
            "input": "echo of input deck",
            "titlecard": "Northwest Computational Chemistry Package (NWChem)",
            "jobinfo": "Job information",
            "meminfo": "Memory information",
            "directoryinfo": "Directory information",
            "dftmodule": "NWChem DFT Module",
            "responsemodule": "*** RESPONSE ***",
            "tddftmodule": "NWChem TDDFT Module",
            "rttddftmodule": "NWChem Real-Time TDDFT Module",
            "gastats": "GA Statistics for process",
        }

        # chunk the output into major sections
        chunk_starts, chunk_types = zip(*self._chunk_file())
        chunks = self._get_chunks(chunk_starts=chunk_starts, chunk_types=chunk_types)

        def get(self, *quantity_names: str):
            method_dict = dict(inspect.getmembers(self, predicate=inspect.isroutine))

            with open(self.filepath, "r") as f:
                lines = "".join(f.readlines())

            quantities = tuple(
                method_dict[name.lower()](lines=lines) for name in quantity_names
            )

            if len(quantities) == 1:
                return quantities[0]
            return quantities

        # fragment each chunk into minor sections
        def modify_chunk_type(x, data={}):
            if x in data:
                data[x] += 1
            else:
                data[x] = 0

            return "{0}{1}".format(x, data[x])

        for chunk, chunk_type in zip(chunks, chunk_types):

            def modify_frag_type(
                x, data={}
            ):  # this is some clever trickery with the default value data={}
                if x in data:
                    data[x] += 1
                else:
                    data[x] = 0
                return "{0}{1}".format(x, data[x])

            fragment_starts, fragment_types = zip(
                *self._fragment_chunk(chunk=chunk, chunk_type=chunk_type)
            )

            unique_fragment_types = [
                modify_frag_type(x=frag_type) for frag_type in fragment_types
            ]
            fragments = self._get_fragments(
                chunk=chunk, fragment_starts=fragment_starts
            )

            self.output_dict[modify_chunk_type(x=chunk_type)] = collections.OrderedDict(
                zip(unique_fragment_types, fragments)
            )

    def _chunk_file(self: str) -> List[Tuple[str, str]]:
        return sorted(
            [
                (x, chunk_type)
                for chunk_type, chunk_keyword in self.chunk_dict.items()
                for x in self._find_chunk_keyword_in_file(chunk_keyword=chunk_keyword)
            ]
        )

    def _find_chunk_keyword_in_file(self, chunk_keyword: str) -> List[int]:
        with open(self.filename, "r") as f:
            return [linenum for linenum, line in enumerate(f) if chunk_keyword in line]

    def _get_chunks(
        self, chunk_starts: Iterable[int], chunk_types: Iterable[str]
    ) -> List[str]:
        cur_ind = 0
        chunk_start = chunk_starts[cur_ind]
        chunk_end = (
            chunk_starts[cur_ind + 1]
            if cur_ind != len(chunk_starts) - 1
            else float("inf")
        )
        cur_chunk = ""
        chunks = []
        with open(self.filename, "r") as f:
            for linenum, line in enumerate(f):
                if linenum < chunk_start:
                    pass
                elif linenum >= chunk_start and linenum < chunk_end:
                    cur_chunk += line
                else:
                    cur_ind += 1
                    chunk_start = chunk_starts[cur_ind]
                    chunk_end = (
                        chunk_starts[cur_ind + 1]
                        if cur_ind != len(chunk_starts) - 1
                        else float("inf")
                    )
                    chunks.append(cur_chunk)
                    cur_chunk = line
        chunks.append(cur_chunk)

        return chunks

    def _fragment_chunk(self, chunk, chunk_type):
        if chunk_type == "input":
            fragment_dict = {"input": "echo of input deck"}
        elif chunk_type == "titlecard":
            fragment_dict = {
                "titlecard": "Northwest Computational Chemistry Package (NWChem)"
            }
        elif chunk_type == "jobinfo":
            fragment_dict = {"jobinfo": "Job information"}
        elif chunk_type == "meminfo":
            fragment_dict = {"meminfo": "Memory information"}
        elif chunk_type == "directoryinfo":
            fragment_dict = {"directoryinfo": "Directory information"}
        elif chunk_type == "dftmodule":
            fragment_dict = {
                "basis": 'Basis "ao basis"',
                "basissummary": 'Summary of "ao basis"',
                "geninfo": "General Information",
                "xcinfo": "XC Information",
                "gridinfo": "Grid Information",
                "convergenceinfo": "Convergence Information",
                "screeninginfo": "Screening Tolerance Information",
                "guess": "Superposition of Atomic Density Guess",
                "scf": "Memory utilization after 1st SCF pass",
                "orbitals": "DFT Final Molecular Orbital Analysis",
                "alpha_orbitals": "DFT Final Alpha Molecular Orbital Analysis",
                "beta_orbitals": "DFT Final Beta Molecular Orbital Analysis",
                "com": "center of mass",
                "moments": "moments of inertia",
                "multipole": "Multipole analysis of the density",
            }
        elif chunk_type == "responsemodule":
            fragment_dict = {
                "header": "Response module for NWChem",
                "field": "Solving response equations for perturbing field",
                "cphf": "NWChem CPHF Module",
                "electricdipoleresponse": "Electric Dipole Response Matrix",
                "linearresponsepolarizability": "DFT Linear Response polarizability",
                "magneticdipoleresponse": "Magnetic Dipole Response Matrix",
            }
        elif chunk_type == "tddftmodule":
            fragment_dict = {
                "geninfo": "General Information",
                "xcinfo": "XC Information",
                "tddftinfo": "TDDFT Information",
                "meminfo": "Memory Information",
                "evaldiffs": "smallest eigenvalue differences",
            }
        elif chunk_type == "rttddftmodule":
            fragment_dict = {
                "basissummary": 'Summary of "ao basis"',
                "geninfo": "General Information",
                "xcinfo": "XC Information",
                "convergenceinfo": "Convergence Information",
                "screeninginfo": "Screening Tolerance Information",
                "propparams": "Propagation parameters",
                "dipolequadrupole": "Dipole and quadrupole matrices",
                "appliedfields": "Applied fields",
                "excitationrules": "Excitation rules",
                "propheader": "shell propagation",
                "propstart": "Propagation started",
                "checkspassed": "Checks passed",
                "savedrestart": "Saved restart file",
                "tddipole": "Dipole moment [system]",
                "istart": "istart",
                "dipcontribx": "Mu-X i-a Contrib",
                "dipcontriby": "Mu-Y i-a Contrib",
                "dipcontribz": "Mu-Z i-a Contrib",
                "propfinish": "Propagation finished",
            }
        elif chunk_type == "gastats":
            fragment_dict = {
                "summary": "Summary of allocated global arrays",
                "stats": "GA Statistics",
            }

        return sorted(
            [
                (x, frag_type)
                for frag_type, frag in fragment_dict.items()
                for x in self._find_fragment_in_chunk(
                    chunk=chunk.split("\n"), fragment=frag
                )
            ]
        )

    def _find_fragment_in_chunk(self, chunk, fragment):
        return [linenum for linenum, line in enumerate(chunk) if fragment in line]

    def _get_fragments(self, chunk, fragment_starts):
        cur_ind = 0
        fragment_start = fragment_starts[cur_ind]
        fragment_end = (
            fragment_starts[cur_ind + 1]
            if cur_ind != len(fragment_starts) - 1
            else float("inf")
        )
        cur_fragment = ""
        fragments = []
        for linenum, line in enumerate(chunk.split("\n")):
            if linenum < fragment_start:
                pass
            elif linenum >= fragment_start and linenum < fragment_end:
                cur_fragment += line + "\n"
            else:
                cur_ind += 1
                fragment_start = fragment_starts[cur_ind]
                fragment_end = (
                    fragment_starts[cur_ind + 1]
                    if cur_ind != len(fragment_starts) - 1
                    else float("inf")
                )
                fragments.append(cur_fragment)
                cur_fragment = line

        fragments.append(cur_fragment)

        return fragments

    def get_orbital_info(
        self, orbitals=None, spin=None, dftmodule_num=0, orbitals_fragment_num=0
    ):
        dft_key = "dftmodule{0}".format(dftmodule_num)

        if spin is None:
            orbital_key = "orbitals{0}".format(orbitals_fragment_num)
        elif hasattr(spin, "lower") and spin.lower() == "alpha":
            orbital_key = "alpha_orbitals{0}".format(orbitals_fragment_num)
        elif hasattr(spin, "lower") and spin.lower() == "beta":
            orbital_key = "beta_orbitals{0}".format(orbitals_fragment_num)
        else:
            raise ValueError(
                "Invalid value for argument spin, which can take values 'alpha','beta', or None."
            )

        orbital_fragment = self.output_dict[dft_key][orbital_key]

        energies = []
        vector_nums = []
        occupancies = []

        mo_centers = []
        mo_r2s = []  # shows how diffuse an orbital is

        for line in orbital_fragment.split("\n"):
            if "Vector" in line:
                partial_1, partial_2, energy = line.split("=")
                _, vector_num, _ = partial_1.split()
                occupancy, _ = partial_2.split()

                energy = float(energy.replace("D", "e"))
                energies.append(energy)

                vector_nums.append(int(vector_num))

                occupancy = float(occupancy.replace("D", "e"))
                occupancies.append(occupancy)
            elif "MO Center" in line:
                _, partial_1, r2 = line.split("=")
                x, y, z, _, = partial_1.strip().split(",")

                x = float(x.replace("D", "e"))
                y = float(y.replace("D", "e"))
                z = float(z.replace("D", "e"))
                mo_centers.append((x, y, z))

                r2 = float(r2.replace("D", "e"))
                mo_r2s.append(r2)

        orbital_dicts = (
            {
                "energy": units.Qty(value=eng, unit=units.hartree),
                "occupancy": units.Qty(value=occ, unit=units.Unit(base="unitless")),
                "center": units.Qty(value=moc, unit=units.Unit(base="bohr")),
                "R2": units.Qty(value=mor2, unit=units.Unit(base="bohr") ** 2),
            }
            for eng, occ, moc, mor2 in zip(energies, occupancies, mo_centers, mo_r2s)
        )

        if orbitals is None:
            return dict(zip(vector_nums, orbital_dicts))
        else:
            return {
                num: od
                for num, od in zip(vector_nums, orbital_dicts)
                if num in orbitals
            }

    def get_homo_lumo(self, dftmodule_num=0, orbitals_fragment_num=0):
        dft_key = "dftmodule{0}".format(dftmodule_num)

        if self.is_spin_polarized(dftmodule_num=dftmodule_num):
            alpha_info = self.get_orbital_info(
                spin="alpha",
                dftmodule_num=dftmodule_num,
                orbitals_fragment_num=orbitals_fragment_num,
            )
            alpha_homo = max(
                od["energy"] for od in alpha_info.values() if od["occupancy"] != 0
            )
            alpha_lumo = min(
                od["energy"] for od in alpha_info.values() if od["occupancy"] == 0
            )

            beta_info = self.get_orbital_info(
                spin="beta",
                dftmodule_num=dftmodule_num,
                orbitals_fragment_num=orbitals_fragment_num,
            )
            beta_homo = max(
                od["energy"] for od in beta_info.values() if od["occupancy"] != 0
            )
            beta_lumo = min(
                od["energy"] for od in beta_info.values() if od["occupancy"] == 0
            )

            homo = max(alpha_homo, beta_homo)
            lumo = min(alpha_lumo, beta_lumo)
        else:
            orbital_info = self.get_orbital_info(
                dftmodule_num=dftmodule_num, orbitals_fragment_num=orbitals_fragment_num
            )
            homo = max(
                [od["energy"] for od in orbital_info.values() if od["occupancy"] != 0]
            )
            lumo = min(
                [od["energy"] for od in orbital_info.values() if od["occupancy"] == 0]
            )

        return {"homo": homo, "lumo": lumo, "gap": lumo - homo}

    def is_spin_polarized(self, dftmodule_num=0):
        dft_key = "dftmodule{0}".format(dftmodule_num)
        return (
            sum(
                (k.startswith("alpha_orbitals") or k.startswith("beta_orbitals"))
                for k in self.output_dict[dft_key].keys()
            )
            > 0
        )

    def get_total_energies(self, dftmodule_num=0, scf_fragment_num=0):
        dft_key = "dftmodule{0}".format(dftmodule_num)
        scf_key = "dftmodule{0}".format(scf_fragment_num)

        energy_fragment = self.output_dict[dft_key][scf_key]

        energy_dict = {}
        for line in energy_fragment.split("\n"):
            if "Total DFT energy" in line:
                _, total_energy = line.split("=")
                total_energy = float(total_energy.replace("D", "e"))
                energy_dict["total_energy"] = units.Qty(
                    value=total_energy, unit=units.hartree
                )
            elif "One electron energy" in line:
                _, one_electron_energy = line.split("=")
                one_electron_energy = float(one_electron_energy.replace("D", "e"))
                energy_dict["one_electron_energy"] = units.Qty(
                    value=one_electron_energy, unit=units.hartree
                )
            elif "Coulomb energy" in line:
                _, coulomb_energy = line.split("=")
                coulomb_energy = float(coulomb_energy.replace("D", "e"))
                energy_dict["coulomb_energy"] = units.Qty(
                    value=coulomb_energy, unit=units.hartree
                )
            elif "Exchange-Corr. energy" in line:
                _, xc_energy = line.split("=")
                xc_energy = float(xc_energy.replace("D", "e"))
                energy_dict["xc_energy"] = units.Qty(
                    value=xc_energy, unit=units.hartree
                )
            elif "Nuclear repulsion energy" in line:
                _, nuclear_repulsion_energy = line.split("=")
                nuclear_repulsion_energy = float(
                    nuclear_repulsion_energy.replace("D", "e")
                )
                energy_dict["nuclear_repulsion_energy"] = units.Qty(
                    value=nuclear_repulsion_energy, unit=units.hartree
                )
            elif "COSMO energy" in line:
                _, cosmo_energy = line.split("=")
                cosmo_energy = float(cosmo_energy.replace("D", "e"))
                energy_dict["cosmo_energy"] = units.Qty(
                    value=cosmo_energy, unit=units.hartree
                )

        return energy_dict

    def get_polarizability(self, responsemodule_num=0, polarize_fragment_num=0):
        response_key = "responsemodule{0}".format(responsemodule_num)
        polarize_key = "linearresponsepolarizability{0}".format(polarize_fragment_num)

        polarizability_fragment = self.output_dict[response_key][polarize_key].split(
            "\n"
        )

        header_string = polarizability_fragment[:4]
        matrix_string = polarizability_fragment[4:7]
        eigenvalues_string = polarizability_fragment[8]
        isotropic_string = polarizability_fragment[9]
        anisotropic_string = polarizability_fragment[10]

        matrix_elements = [
            float(str)
            for m in matrix_string
            for str in m.split()
            if str not in ("X", "Y", "Z")
        ]
        polarizability_matrix = np.array(matrix_elements).reshape((3, 3))

        _, _, eig1, eig2, eig3 = eigenvalues_string.split()
        eigenvalues = (float(eig1), float(eig2), float(eig3))

        _, _, iso = isotropic_string.split()
        isotropic_polarizability = float(iso)

        _, _, aniso = anisotropic_string.split()
        anisotropy = float(aniso)

        return units.Qty(value=polarizability_matrix, unit=units.au_volume)

        # return {'polarizability_matrix': units.Qty(value=polarizability_matrix,unit=units.au_volume),
        #         'polarizability_eigenvalues': units.Qty(value=eigenvalues,unit=units.au_volume),
        #         'isotropic_polarizability': units.Qty(value=isotropic_polarizability,unit=units.au_volume),
        #         'polarizability_anisotropy': units.Qty(value=anisotropy,unit=units.au_volume),
        #        }

    def get_tddipole(self, rttddftmodule_num=None, field_tag=None, appliedfields_num=0):
        if rttddftmodule_num is None:
            rttddft_keys = (
                k for k in self.output_dict.keys() if k.startswith("rttddftmodule")
            )
        else:
            rttddft_keys = (f"rttddftmodule{rttddftmodule_num}",)

        result_dict = {}

        for rttddft_key in rttddft_keys:
            rttddft_dict = self.output_dict[rttddft_key]

            tddipole_keys = [k for k in rttddft_dict.keys() if k.startswith("tddipole")]

            if field_tag is None:
                field_tags = (
                    tag
                    for tag, _, _, _, _, _, _, _, _ in (
                        rttddft_dict[k].split() for k in tddipole_keys
                    )
                )
            else:
                field_tags = (field_tag,)

            result = {}

            for ft in field_tags:
                tddipole_keys = [
                    k for k in rttddft_dict.keys() if k.startswith("tddipole")
                ]
                time, mu_x, mu_y, mu_z = zip(
                    *(
                        (float(t), float(x), float(y), float(z))
                        for tag, t, x, y, z, _, _, _, _ in (
                            rttddft_dict[k].split() for k in tddipole_keys
                        )
                        if tag == ft
                    )
                )

                applied_field_fragment = rttddft_dict[
                    f"appliedfields{appliedfields_num}"
                ].split("\n")
                applied_field_header = applied_field_fragment[:3]
                (
                    applied_field_type_str,
                    applied_field_polarization_str,
                    applied_field_maximum_str,
                    _,
                    _,
                    _,
                ) = applied_field_fragment[3:]

                _, _, applied_field_type = applied_field_type_str.split()
                (
                    _,
                    _,
                    applied_field_polarization,
                ) = applied_field_polarization_str.split()
                (
                    _,
                    _,
                    _,
                    _,
                    _,
                    _,
                    applied_field_maximum,
                    _,
                ) = applied_field_maximum_str.split()

                time_unit = units.Unit("au_time")
                dipole_unit = units.au_dipole_moment
                applied_field_unit = units.volt / units.Unit("nanometer")

                result[ft] = {
                    "time": units.Qty(value=np.array(time), unit=time_unit),
                    "mu_x": units.Qty(value=np.array(mu_x), unit=dipole_unit),
                    "mu_y": units.Qty(value=np.array(mu_y), unit=dipole_unit),
                    "mu_z": units.Qty(value=np.array(mu_z), unit=dipole_unit),
                    "E_max": units.Qty(
                        value=applied_field_maximum, unit=applied_field_unit
                    ),
                    "E_shape": applied_field_type,
                    "E_polarization": applied_field_polarization,
                }

            result_dict[rttddft_key] = result

        return result_dict

    def plot_tddipole(self, rttddftmodule_num=None, field_tag=None):
        tddipole_dict = self.get_tddipole(
            field_tag=field_tag, rttddftmodule_num=rttddftmodule_num
        )
        for rttddftmodule_num, result_dict in tddipole_dict.items():
            for field_tag, result in result_dict.items():
                time = result["time"]
                mu_x = result["mu_x"]
                mu_y = result["mu_y"]
                mu_z = result["mu_z"]

                plt.plot(
                    time.value, mu_x.value, label=f"x_{field_tag}_{rttddftmodule_num}"
                )
                plt.plot(
                    time.value, mu_y.value, label=f"y_{field_tag}_{rttddftmodule_num}"
                )
                plt.plot(
                    time.value, mu_z.value, label=f"z_{field_tag}_{rttddftmodule_num}"
                )
                plt.title("Dynamic dipole moments")
                plt.xlabel("Time ({0})".format(time.unit))
                plt.ylabel("Dipole moment ({0})".format(mu_x.unit))

        plt.legend()
        plt.show()

    def get_dipcontribs(self, dipcontrib_dir, rttddftmodule_num=0, i=None, a=None):
        # FIXME: there is currently no way to get dipcontribs by field tag like
        # there is with get_tddipole; consider fixing by not fragmenting dipcontribs
        # and tddipoles apart but rather storing them together
        rttddft_key = "rttddftmodule{0}".format(rttddftmodule_num)
        rttddft_dict = self.output_dict[rttddft_key]

        # dipcontrib_keys = [k for k in rttddft_dict.keys() if k.startswith('dipcontrib')]
        # dipcontribx_keys = [k for k in dipcontrib_keys if k.startswith('dipcontribx')]
        # dipcontriby_keys = [k for k in dipcontrib_keys if k.startswith('dipcontriby')]
        # dipcontribz_keys = [k for k in dipcontrib_keys if k.startswith('dipcontribz')]
        #
        # for direction in ('x','y','z'):
        #     dipcontrib_dir_keys = [k for k in dipcontrib_keys if k.startswith('dipcontrib{0}'.format(direction))]
        #     from_orbitals_key,to_orbitals_key,propagation_keys = dipcontrib_dir_keys[0],dipcontrib_dir_keys[1],dipcontrib_dir_keys[2:]
        #
        #     orbital_pairs = [(int(i),int(a)) for i,a in zip(rttddft_dict[from_orbitals_key].split(),rttddft_dict[to_orbitals_key].split()) if i not in ('#Mu-X','#Mu-Y','#Mu-Z','i-a','Contrib','CLOSED') and a not in ('#Mu-X','#Mu-Y','#Mu-Z','i-a','Contrib','CLOSED')]
        #
        #     np.array([[float(p) for p in rttddft_dict[k].split()[:-4]] for k in propagation_keys]))
        #
        #
        #
        # from_orbital_nums = tuple(int(i) for i in rttddft_dict[dipcontrib_keys[0]].split()[:-4])
        # to_orbital_nums = tuple(int(a) for a in rttddft_dict[dipcontrib_keys[1]].split()[:-4])
        #
        # zip(*((float(t),float(x),float(y),float(z)) for tag,t,x,y,z,_,_,_,_ in (rttddft_dict[k].split() for k in propagation_keys) if tag == field_tag))
        #
        # print(len(rttddft_dict[dipcontrib_keys[2]].split()[:-4]))
        #
        # print(len(list(zip(from_orbital_nums,to_orbital_nums))))

        dipcontribs = [
            [float(x) for x in rttddft_dict[k].split()[:-4]]
            for k in rttddft_dict.keys()
            if "dipcontrib{0}".format(dipcontrib_dir) in k
        ]
        # need to cut out first two rows (don't currently understand these outputs from NWChem)
        # first column (first row after transpose) represents the time
        ia_pairs = np.array([(int(x), int(y)) for x, y in zip(*dipcontribs[0:2])])
        dipcontrib_array = np.array(dipcontribs[2:]).T
        time = dipcontrib_array[0]

        return {
            "time": units.Qty(value=time, unit=units.Unit(base="au_time")),
            "dipole_contributions": units.Qty(
                value=dipcontribs, unit=units.au_dipole_moment
            ),
            "orbital_pairs": units.Qty(
                value=ia_pairs, unit=units.Unit(base="unitless")
            ),
        }

        # if i is not None or a is not None:
        #     if i is not None and a is None:
        #         plot_inds = [ind+1 for ind,(j,b) in enumerate(ia_pairs) if j == i]
        #     elif i is None and a is not None:
        #         plot_inds = [ind+1 for ind,(j,b) in enumerate(ia_pairs) if b == a]
        #     elif i is not None and a is not None:
        #         plot_inds = [ind+1 for ind,(j,b) in enumerate(ia_pairs) if j == i and b == a]
        #
        #     ia_pair_inds = [ind-1 for ind in plot_inds]
        #
        #     return ia_pairs[ia_pair_inds],time,dipcontrib_array[plot_inds]
        # else:

    #     #     return ia_pairs,time,dipcontrib_array[1:]
    #
    # def plot_dipcontrib(self,rttddft_key,dipcontrib_dir,i=None,a=None):
    #     ia_pairs,time,dipcontrib_array = self.get_dipcontribs(rttddft_key=rttddft_key,dipcontrib_dir=dipcontrib_dir)
    #
    #     if i is not None and a is None:
    #         ia_pair_inds = [ind for ind,(j,b) in enumerate(ia_pairs) if j == i]
    #     elif i is None and a is not None:
    #         ia_pair_inds = [ind for ind,(j,b) in enumerate(ia_pairs) if b == a]
    #     elif i is not None and a is not None:
    #         ia_pair_inds = [ind for ind,(j,b) in enumerate(ia_pairs) if j == i and b == a]
    #     else:
    #         ia_pair_inds = [ind for ind,(j,b) in enumerate(ia_pairs)]
    #
    #     #!!! the * notation here is not correct yet - some occupied orbitals are included!!
    #     for row in dipcontrib_array[ia_pair_inds]:
    #         plt.plot(time,row)
    #     plt.legend(['{0} --> {1}*'.format(x,y) for x,y in ia_pairs[ia_pair_inds]])
    #     plt.show()
    #
    #     plt.plot(-0.5*np.sum(dipcontrib_array,0))


# def parse_dipole(filename):
#     dipole_dict = {}
#     with open(filename,'r') as f:
#         dmx_lines = [line for line in f if 'DMX' in line]
#     with open(filename,'r') as f:
#         dmy_lines = [line for line in f if 'DMY' in line]
#     with open(filename,'r') as f:
#         dmz_lines = [line for line in f if 'DMZ' in line]
#
#     dipole_dict['dmx_au'] = float(dmx_lines[0].split()[1])
#     dipole_dict['dmx_efc_au'] = float(dmx_lines[0].split()[-1])
#     dipole_dict['dmx_debye'] = float(dmx_lines[1].split()[1])
#     dipole_dict['dmx_efc_debye'] = float(dmx_lines[1].split()[-1])
#
#     dipole_dict['dmy_au'] = float(dmy_lines[0].split()[1])
#     dipole_dict['dmy_efc_au'] = float(dmy_lines[0].split()[-1])
#     dipole_dict['dmy_debye'] = float(dmy_lines[1].split()[1])
#     dipole_dict['dmy_efc_debye'] = float(dmy_lines[1].split()[-1])
#
#     dipole_dict['dmz_au'] = float(dmz_lines[0].split()[1])
#     dipole_dict['dmz_efc_au'] = float(dmz_lines[0].split()[-1])
#     dipole_dict['dmz_debye'] = float(dmz_lines[1].split()[1])
#     dipole_dict['dmz_efc_debye'] = float(dmz_lines[1].split()[-1])
#
#     dipole_dict['total_dipole_au'] = np.linalg.norm([dipole_dict['dmx_au']+dipole_dict['dmx_efc_au'],dipole_dict['dmy_au']+dipole_dict['dmy_efc_au'],dipole_dict['dmz_au']+dipole_dict['dmz_efc_au']])
#     dipole_dict['total_dipole_debye'] = np.linalg.norm([dipole_dict['dmx_debye']+dipole_dict['dmx_efc_debye'],dipole_dict['dmy_debye']+dipole_dict['dmy_efc_debye'],dipole_dict['dmz_debye']+dipole_dict['dmz_efc_debye']])
#
#     return dipole_dict



from __future__ import annotations
import cclib
import inspect
import numpy as np
import materia as mtr
import re
from typing import Dict, Tuple, TypeVar, Union

__all__ = ["QChemOutput"]

T = TypeVar("T")


class QChemOutput:
    def __init__(self, filepath: str) -> None:
        self.filepath = mtr.expand(filepath)
        self.cclib_out = cclib.io.ccread(self.filepath)

    def get(self, *quantity_names: str) -> Union[T, Tuple[T]]:
        method_dict = dict(inspect.getmembers(self, predicate=inspect.isroutine))

        with open(self.filepath, "r") as f:
            lines = "".join(f.readlines())

        quantities = tuple(
            method_dict[name.lower()](lines=lines) for name in quantity_names
        )

        if len(quantities) == 1:
            return quantities[0]
        return quantities

    def footer(
        self, lines: str
    ):  # FIXME: not sure how to annotate type hint: is it -> Dict[mtr.Qty,mtr.Qty,Tuple[int,str,int,int,int,int,str]] ?
        s = r"\s*Total\s*job\s*time\s*:\s*(\d*\.\d*)\s*s\s*\(\s*wall\s*\)\s*,\s*(\d*\.\d*)\s*s\s*\(\s*cpu\s*\)\s*(\w*)\s*(\w*)\s*(\d*)\s*(\d*)\s*:\s*(\d*)\s*:\s*(\d*)\s*(\d*)\s*"
        pattern = re.compile(s)
        (
            walltime,
            cputime,
            day,
            month,
            date,
            hour,
            minutes,
            seconds,
            year,
        ) = pattern.search(lines).groups()

        walltime = float(walltime) * mtr.second
        cputime = float(cputime) * mtr.second
        date = tuple(
            int(year), month, int(date), int(hour), int(minutes), int(seconds), day
        )

        return {"walltime": walltime, "cputime": cputime, "date": date}

    def kohn_sham_gap(self, lines: str):  # FIXME: not sure how to annotate type hint
        s = (
            r"\s*---\s*Generalized\s*Kohn-Sham\s*gap\s*\((\S*)\)\s*---"
            r"\s*HOMO\s*Eigenvalue\s*=\s*(-?\d*\.*\d*)\s*LUMO\s*Eigenvalue\s*"
            r"=\s*(-?\d*\.*\d*)\s*KS\s*gap\s*=\s*(-?\d*\.*\d*)\s*"
        )

        energy_unit_str, homo, lumo, gap = re.search(s, lines).groups()

        if energy_unit_str.lower() == "ev":
            unit = mtr.eV
        else:
            raise ValueError("Cannot parse energy unit.")

        homo = float(homo) * unit
        lumo = float(lumo) * unit
        gap = float(gap) * unit

        return {"homo": homo, "lumo": lumo, "gap": gap}

    def polarizability(self, lines: str) -> mtr.Qty:
        s = (
            r"\s*Polarizability tensor\s*\[(.*)\]\s*(-?\d*\.\d*)\s*"
            r"(-?\d*\.\d*)\s*(-?\d*\.\d*)\s*(-?\d*\.\d*)\s*(-?\d*\.\d*)\s*"
            r"(-?\d*\.\d*)\s*(-?\d*\.\d*)\s*(-?\d*\.\d*)\s*(-?\d*\.\d*)\s*"
        )
        pattern = re.compile(s)
        # FIXME: more robust (or any) error checking/edge-case handling needed here
        unit_str, *tensor_components = pattern.search(lines).groups()

        if unit_str.lower() == "a.u.":
            unit = mtr.au_volume

        polarizability_tensor = (
            np.array(tuple(float(tc) for tc in tensor_components)).reshape(3, 3) * unit
        )

        return mtr.Polarizability(polarizability_tensor=polarizability_tensor)

    def rttddft(self, lines: str):  # FIXME: not sure how to annotate type hint
        s = r"ITER:\s*(\d*)\s*T:\s*(\d*\.\d*)\(fs\)\s*dt\s*(\d*\.\d*)\(fs\)\s*Hr/Ps:\s*(\d*\.\d*)\s*-\s*Lpsd/Rem\.:\s*(\d*\.\d*),\s*([^\s]*)\s*\(min\)\s*Tr\.Dev:\s*(\d*\.\d*)\s*Hrm:\s*(\d*\.\d*)\s*Enrgy:\s*(-?\d*\.\d*)\s*Entr:\s*(-?\d*.\d*)\s*Fld\s*(\d*)\s*NFk:\s*(\d*)\s*Mu\s*(-?\d*\.\d*e?-?\d*)\s*(-?\d*\.\d*e?-?\d*)\s*(-?\d*\.\d*e?-?\d*)"
        pattern = re.compile(s)

        iterations = []
        Ts = []
        dts = []
        hours_per_ps = []
        lapsed = []  # FIXME: what is this?
        remaining = []  # FIXME: what is this?
        trace_deviations = []  # FIXME: what is this?
        hrms = []  # FIXME: what is this?
        energies = []
        entrs = []  # FIXME: what is this?
        field = []  # FIXME: what is this? field on or off, always a boolean?
        number_fock = []
        mu_xs = []
        mu_ys = []
        mu_zs = []

        for (
            iter,
            T,
            dt,
            hps,
            lpsd,
            rem,
            tr_dev,
            hrm,
            energy,
            entr,
            fld,
            nfk,
            mu_x,
            mu_y,
            mu_z,
        ) in pattern.findall(lines):
            iterations.append(int(iter))
            Ts.append(float(T))
            dts.append(float(dt))
            hours_per_ps.append(float(hps))
            lapsed.append(float(lpsd))
            remaining.append(float(rem))
            trace_deviations.append(float(tr_dev))
            hrms.append(float(hrm))
            energies.append(float(energy))
            entrs.append(float(entr))
            field.append(int(fld))
            number_fock.append(int(nfk))
            mu_xs.append(float(mu_x))
            mu_ys.append(float(mu_y))
            mu_zs.append(float(mu_z))

        return {
            "iterations": iterations,
            "T": np.array(Ts) * mtr.fs,
            "dt": np.array(dts) * mtr.fs,
            "hours_per_picosecond": np.array(hours_per_ps) * mtr.hr / mtr.ps,
            "lapsed": np.array(lapsed) * mtr.minute,
            "remaining": np.array(remaining) * mtr.minute,
            "trace_deviations": trace_deviations,
            "hrm": hrms,
            "energies": np.array(energies) * mtr.hartree,
            "entr": entrs,
            "field": field,
            "number_fock": number_fock,
            "mu_x": np.array(mu_xs) * mtr.au_dipole_moment,
            "mu_y": np.array(mu_ys) * mtr.au_dipole_moment,
            "mu_z": np.array(mu_zs) * mtr.au_dipole_moment,
        }

    def electronic_excitations(
        self, lines: str
    ):  # FIXME: not sure how to annotate type hint
        engs = mtr.h * mtr.c * (self.cclib_out.etenergies / mtr.cm)
        engs = engs.convert(mtr.eV)
        excitations = tuple(
            mtr.Excitation(
                energy=eng, oscillator_strength=osc, symmetry=sym, contributions=cont
            )
            for eng, osc, sym, cont in zip(
                engs,
                self.cclib_out.etoscs,
                self.cclib_out.etsyms,
                self.cclib_out.etsecs,
            )
        )
        return mtr.ExcitationSpectrum(excitations)

    def _process_ee_block(
        self, block, pattern_1, pattern_2
    ):  # FIXME: not sure how to annotate type hint
        (
            excitation_energy_unit_str,
            excitation_energy,
            total_energy,
            total_energy_unit_str,
            multiplicity,
            moment_x,
            moment_y,
            moment_z,
            strength_str,
            *excitation_strs,
        ) = pattern_1.search(block).groups()

        if excitation_energy_unit_str.lower() == "ev":
            excitation_energy_unit = mtr.eV
        excitation_energy = float(excitation_energy) * excitation_energy_unit

        if total_energy_unit_str.lower() == "au":
            total_energy_unit = mtr.au_energy
        total_energy = float(total_energy) * total_energy_unit

        if multiplicity.lower() == "singlet":
            mult = 1
        elif multiplicity.lower() == "triplet":
            mult = 3

        moment = np.array((float(moment_x), float(moment_y), float(moment_z)))
        strength = float(strength_str)

        contributions = []
        for s in excitation_strs:
            from_occ, from_num, to_occ, to_num, amplitude = pattern_2.search(s).groups()
            contributions.append(
                (from_occ, int(from_num), to_occ, int(to_num), float(amplitude))
            )

        return mtr.ElectronicExcitation(
            excitation_energy=excitation_energy,
            total_energy=total_energy,
            multiplicity=mult,
            transition_moment=moment,
            oscillator_strength=strength,
            contributions=contributions,
        )

    def orbital_energies(self, lines: str) -> Tuple[Tuple[mtr.Qty]]:
        energy_symmetry_pattern = re.compile(
            r"((?:-?\d*\.\d*\s*)*(?:\d*\s*[a-zA-Z]*\d*\s*)*)"
        )
        energy_pattern = re.compile(r"(-?\d*\.\d*)")
        symmetry_pattern = re.compile(r"(\d*\s*[a-zA-Z]+\d*)")

        s = r"\s*Orbital\s*Energies\s*\((a\.u\.)\)\s*and\s*Symmetries"
        pattern = re.compile(s)
        (energy_unit_str,) = pattern.search(lines).groups()
        if energy_unit_str == "a.u.":
            energy_unit = mtr.hartree
        else:
            raise ValueError("Cannot parse energy unit in Orbital Energies section.")

        s = (
            r"\s*Alpha\s*MOs.*?"
            r"\s*--\s*Occupied\s*--\s*((?:(?:-?\d*\.\d*\s*)*(?:\d*\s*[a-zA-Z]*\d*\s*)*)*)"
            r"\s*--\s*Virtual\s*--\s*((?:(?:-?\d*\.\d*)*\s*(?:\d*\s*[a-zA-Z]*\d*)*\s*)*)"
        )
        pattern_alpha = re.compile(s)
        alpha_orbitals_occupied, alpha_orbitals_virtual = pattern_alpha.search(
            lines
        ).groups()

        alpha_orbital_occupied_energies = tuple(
            float(x) * energy_unit
            for split in energy_symmetry_pattern.findall(alpha_orbitals_occupied)
            for x in energy_pattern.findall(split)
            if split != ""
        )
        alpha_orbital_virtual_energies = tuple(
            float(x) * energy_unit
            for split in energy_symmetry_pattern.findall(alpha_orbitals_virtual)
            for x in energy_pattern.findall(split)
            if split != ""
        )
        # alpha_orbital_occupied_symmetries = tuple(x for split in energy_symmetry_pattern.findall(alpha_orbitals_occupied) for x in symmetry_pattern.findall(split) if split != '')

        s = (
            r"\s*Beta\s*MOs.*?"
            r"\s*--\s*Occupied\s*--\s*((?:(?:-?\d*\.\d*\s*)*(?:\d*\s*[a-zA-Z]*\d*\s*)*)*)"
            r"\s*--\s*Virtual\s*--\s*((?:(?:-?\d*\.\d*)*\s*(?:\d*\s*[a-zA-Z]*\d*)*\s*)*)"
        )
        pattern_beta = re.compile(s)
        beta_orbitals_occupied, beta_orbitals_virtual = pattern_beta.search(
            lines
        ).groups()

        beta_orbital_occupied_energies = tuple(
            float(x) * energy_unit
            for split in energy_symmetry_pattern.findall(beta_orbitals_occupied)
            for x in energy_pattern.findall(split)
            if split != ""
        )
        beta_orbital_virtual_energies = tuple(
            float(x) * energy_unit
            for split in energy_symmetry_pattern.findall(beta_orbitals_virtual)
            for x in energy_pattern.findall(split)
            if split != ""
        )
        # beta_orbital_occupied_symmetries = tuple(x for split in energy_symmetry_pattern.findall(beta_orbitals_occupied) for x in symmetry_pattern.findall(split) if split != '')

        return (
            alpha_orbital_occupied_energies,
            beta_orbital_occupied_energies,
            alpha_orbital_virtual_energies,
            beta_orbital_virtual_energies,
        )

    def scf_energy(self, lines: str) -> mtr.Qty:
        s = r"\s*SCF\s*energy\s*in\s*the\s*final\s*basis\s*set\s*=\s*(-?\d*\.\d*)\s*"
        # FIXME: this doesn't seem like a great way to handle multiple groups... better fix?
        *_, energy_str = re.search(s, lines).groups()

        return float(energy_str) * mtr.hartree

    def total_energy(self, lines: str) -> mtr.Qty:
        s = r"\s*Total\s*energy\s*in\s*the\s*final\s*basis\s*set\s*=\s*(-?\d*\.\d*)\s*"
        # FIXME: this doesn't seem like a great way to handle multiple groups... better fix?
        *_, energy_str = re.search(s, lines).groups()

        return float(energy_str) * mtr.hartree


# class Section:
#     def __init__(self, name):
#         self.name = name
#
#     def parse(self, line):
#         pass
#
#     @property
#     def sections(self):
#         section_list = []
#         lines = []
#         section_class = None
#
#         for line in self.lines:
#             new_section_class = next((sc for skw,sc in self.section_keywords.items() if skw in line), None)
#             if new_section_class is None:
#                 lines.append(line)
#             else:
#                 if section_class is not None:
#                     section_list.append(section_class(lines=lines))
#                 section_class = new_section_class
#                 lines = [line]
#
#         section_list.append(section_class(lines=lines))
#
#         return tuple(section_list)
#
#     # def __str__(self):
#     #     return ''.join(self.lines)
#
# class QChemOutput(Section):
#     def __init__(self, filename):
#         self.filename = filename
#         #self.output_dict = collections.OrderedDict()
#         self.job_list = []
#
#     @property
#     def jobs(self):
#         with open(self.filename,'r') as f:
#             for line in f:
#                 self.parse(line)
#
#         return tuple(self.job_list)
#
#     def parse(self, line):
#         if 'Running Job' in line:
#             self.job_list.append(Job())
#         if len(self.job_list) > 0:
#             self.job_list[-1].parse(line=line)
#
# class Job:
#     def __init__(self):
#         self.sections = collections.OrderedDict()
#         self.names = collections.defaultdict(int)
#
#     def parse(self, line):
#         if 'Archival summary:' in line:
#             self.sections[self.append_counts('Footer')] = Footer()
#         elif 'Welcome to Q-Chem' in line:#elif 'Thank you very much for using Q-Chem.  Have a nice day.' in line:
#             self.sections[self.append_counts('Header')] = Header()
#         elif 'Checking the input file for inconsistencies...' in line:
#             self.sections[self.append_counts('InputEcho')] = InputEcho()
#         elif 'TDDFT/TDA calculation will be performed' in line:
#             self.sections[self.append_counts('TDDFTTDARoots')] = TDDFTTDARoots()
#         elif 'Excited state' in line:
#             self.sections[self.append_counts('TDDFTExcitation')] = TDDFTExcitation()
#         # elif 'TDDFT/TDA Excitation Energies' in line:
#         #     self.section_list.append(TDDFTExcitationEnergies())
#         elif 'Orbital Energies' in line:
#             self.sections[self.append_counts('OrbitalEnergies')] = OrbitalEnergies()
#         elif 'Generalized Kohn-Sham gap' in line:
#             self.sections[self.append_counts('KohnShamGap')] = KohnShamGap()
#         elif 'Ground-State Mulliken Net Atomic Charges' in line:
#             self.sections[self.append_counts('NetAtomicCharges')] = NetAtomicCharges()
#         elif 'Cartesian Multipole Moments' in line:
#             self.sections[self.append_counts('CartesianMultipoleMoments')] = CartesianMultipoleMoments()
#         elif 'SETman timing' in line:
#             self.sections[self.append_counts('SETManTiming')] = SETManTiming()
#         elif 'General SCF calculation program' in line:
#             self.sections[self.append_counts('SCF')] = SCF()
#         elif 'Gradient of SCF Energy' in line:
#             self.sections[self.append_counts('SCFGradient')] = SCFGradient()
#         elif 'Standard Nuclear Orientation' in line:
#             self.sections[self.append_counts('StandardNuclearOrientation')] = StandardNuclearOrientation()
#         elif 'Calculating the polarizability using responseman/libresponse' in line:
#             self.sections[self.append_counts('Polarizability')] = Polarizability()
#
#         if len(self.sections) > 0:
#             next(reversed(self.sections.values())).parse(line=line)
#
#     def append_counts(self, name):
#         self.names[name] += 1
#         return f'{name}_{self.names[name]}'
#         #data[x] += 1
#         #
#         #return f'{x}_{data[x]}'
#
# class CartesianMultipoleMoments(Section):
#     def __init__(self):
#         super().__init__(name='CartesianMultipoleMoments')
#         self.charge_next = False
#
#         self.charge_unit = None
#         self.charge_multiplier = None
#         self.charge = None
#         self.dipole_unit = None
#         self.total_dipole = None
#         self.quadrupole_unit = None
#         self.quadrupole = None
#
#     def todict(self):
#         dipole_value = np.array([self.x,self.y,self.z])
#         dipole = mtr.Qty(value=dipole_value,unit=self.dipole_unit)
#
#         quad_value = np.array([self.xx,self.xy,self.xz,self.yx,self.yy,self.yz,self.zx,self.zy,self.zz]).reshape(3,3)
#         quadrupole = mtr.Qty(value=quad_value,unit=self.quadrupole_unit)
#
#         oct_value = np.array([self.xxx,self.xxy,self.xxz,self.xyx,self.xyy,self.xyz,
#                               self.xzx,self.xzy,self.xzz,self.yxx,self.yxy,self.yxz,
#                               self.yyx,self.yyy,self.yyz,self.yzx,self.yzy,self.yzz,
#                               self.zxx,self.zxy,self.zxz,self.zyx,self.zyy,self.zyz,
#                               self.zzx,self.zzy,self.zzz]).reshape(3,3,3)
#         octopole = mtr.Qty(value=oct_value,unit=self.octopole_unit)
#
#         hex_value = np.array([self.xxxx,self.xxxy,self.xxxz,self.xxyx,self.xxyy,
#                               self.xxyz,self.xxzx,self.xxzy,self.xxzz,self.xyxx,
#                               self.xyxy,self.xyxz,self.xyyx,self.xyyy,self.xyyz,
#                               self.xyzx,self.xyzy,self.xyzz,self.xzxx,self.xzxy,
#                               self.xzxz,self.xzyx,self.xzyy,self.xzyz,self.xzzx,
#                               self.xzzy,self.xzzz,self.yxxx,self.yxxy,self.yxxz,
#                               self.yxyx,self.yxyy,self.yxyz,self.yxzx,self.yxzy,
#                               self.yxzz,self.yyxx,self.yyxy,self.yyxz,self.yyyx,
#                               self.yyyy,self.yyyz,self.yyzx,self.yyzy,self.yyzz,
#                               self.yzxx,self.yzxy,self.yzxz,self.yzyx,self.yzyy,
#                               self.yzyz,self.yzzx,self.yzzy,self.yzzz,self.zxxx,
#                               self.zxxy,self.zxxz,self.zxyx,self.zxyy,self.zxyz,
#                               self.zxzx,self.zxzy,self.zxzz,self.zyxx,self.zyxy,
#                               self.zyxz,self.zyyx,self.zyyy,self.zyyz,self.zyzx,
#                               self.zyzy,self.zyzz,self.zzxx,self.zzxy,self.zzxz,
#                               self.zzyx,self.zzyy,self.zzyz,self.zzzx,self.zzzy,
#                               self.zzzz]).reshape(3,3,3,3)
#         hexadecapole = mtr.Qty(value=hex_value,unit=self.hexadecapole_unit)
#
#         return {'charge': self.charge,'dipole': dipole,'total_dipole': self.total_dipole,
#                 'quadrupole': quadrupole,'octopole': octopole,'hexadecapole': hexadecapole}
#
#     def parse(self, line):
#         if 'Charge' in line:
#             self._parse_charge_unit_and_multiplier(line=line)
#             self.charge_next = True
#         elif self.charge_next:
#             self._parse_charge(line=line)
#             self.charge_next = False
#         elif 'Tot ' in line:
#             self._parse_total_dipole(line=line)
#         elif 'Quadrupole Moments' in line:
#             self._parse_quadrupole_unit(line=line)
#         elif 'Octopole Moments' in line:
#             self._parse_octopole_unit(line=line)
#         elif 'Hexadecapole Moments' in line:
#             self._parse_hexadecapole_unit(line=line)
#         elif 'Dipole Moment' in line:
#             self._parse_dipole_unit(line=line)
#         elif 'X' in line or 'Y' in line or 'Z' in line:
#             self._parse_multipole_elements(line=line)
#         else:
#             pass
#
#     def _parse_charge(self, line):
#         charge_value = float(line.strip())*self.charge_multiplier
#         self.charge = mtr.Qty(value=charge_value,unit=self.charge_unit)
#
#     def _parse_charge_unit_and_multiplier(self, line):
#         charge_unit_str,charge_prefactor_base,charge_prefactor_exp = re.search(r'.*\((.*)\s*x\s*(\d*)\^(\d*).*\).*',line).groups()
#
#         charge_unit_str = charge_unit_str.strip().lower()
#         if charge_unit_str == 'esu':
#             self.charge_unit = mtr.esu
#         else:
#             raise ValueError('Cannot parse charge unit in Cartesian Multipole Moments section.')
#
#         self.charge_multiplier = float(charge_prefactor_base)**float(charge_prefactor_exp)
#
#     def _parse_dipole_unit(self, line):
#         (dipole_unit_str,) = re.search(r'.*\((.*).*\).*',line).groups()
#         dipole_unit_str = dipole_unit_str.strip().lower()
#         if dipole_unit_str == 'debye':
#             self.dipole_unit = mtr.debye
#         else:
#             raise ValueError('Cannot parse dipole unit in Cartesian Multipole Moments section.')
#
#     def _parse_quadrupole_unit(self, line):
#         (quad_unit_str,) = re.search(r'.*\((.*).*\).*',line).groups()
#         quad_unit_str = quad_unit_str.strip().lower()
#         if quad_unit_str == 'debye-ang':
#             self.quadrupole_unit = mtr.debye*mtr.angstrom
#         else:
#             raise ValueError('Cannot parse quadrupole unit in Cartesian Multipole Moments section.')
#
#     def _parse_octopole_unit(self, line):
#         (oct_unit_str,) = re.search(r'.*\((.*).*\).*',line).groups()
#         oct_unit_str = oct_unit_str.strip().lower()
#         if oct_unit_str == 'debye-ang^2':
#             self.octopole_unit = mtr.debye*mtr.angstrom**2
#         else:
#             raise ValueError('Cannot parse octopole unit in Cartesian Multipole Moments section.')
#
#     def _parse_hexadecapole_unit(self, line):
#         (hex_unit_str,) = re.search(r'.*\((.*).*\).*',line).groups()
#         hex_unit_str = hex_unit_str.strip().lower()
#         if hex_unit_str == 'debye-ang^3':
#             self.hexadecapole_unit = mtr.debye*mtr.angstrom**3
#         else:
#             raise ValueError('Cannot parse hexadecapole unit in Cartesian Multipole Moments section.')
#
#     def _parse_total_dipole(self, line):
#         _,dipole_total_value = line.split()
#         self.total_dipole = mtr.Qty(value=float(dipole_total_value),unit=self.dipole_unit)
#
#     def _parse_multipole_elements(self, line):
#         for multipole_element_symbol,multipole_element_value in self._pairwise(iterable=line.split()):
#             self._store_multipole_element(element_symbol=multipole_element_symbol.lower(),value=float(multipole_element_value))
#
#     def _pairwise(self, iterable):
#         args = [iter(iterable)] * 2
#         return itertools.zip_longest(*args)
#
#     def _store_multipole_element(self, element_symbol, value):
#         for p in itertools.permutations(element_symbol):
#             setattr(self,''.join(p),value)
#
# class Footer(Section):
#     def __init__(self):
#         super().__init__(name='Footer')
#         self.wall_time = None
#         self.cpu_time = None
#
#     def todict(self):
#         return {'wall_time': self.wall_time, 'cpu_time': self.cpu_time}
#
#     def parse(self, line):
#         if 'Total job time' in line:
#             self._parse_job_time(line=line)
#
#     def _parse_job_time(self, line):
#         wall_time,cpu_time = re.search(r'.*?(\d*\.\d*)s.*?(\d*\.\d*)s.*$',line).groups()
#         self.wall_time = mtr.Qty(value=float(wall_time),unit=mtr.second)
#         self.cpu_time = mtr.Qty(value=float(cpu_time),unit=mtr.second)
#
# class Header(Section):
#     def __init__(self):
#         super().__init__(name='Header')
#
# class InputEcho(Section):
#     def __init__(self):
#         super().__init__(name='InputEcho')
#
# class KohnShamGap(Section):
#     def __init__(self):
#         super().__init__(name='KohnShamGap')
#         self.parse_lookup = {1: self._parse_unit, 2: self._parse_homo, 3: self._parse_lumo, 4: self._parse_gap}
#         self.num_lines = 0
#
#     def parse(self, line):
#         self.num_lines += 1
#         try:
#             self.parse_lookup[self.num_lines](line=line)
#             return True
#         except KeyError:
#             return False
#
#     def _parse_unit(self, line):
#         (unit_string,) = re.search(r'.*\((.*)\).*$',line).groups()
#         if unit_string == 'eV':
#             self.unit = mtr.eV
#         else:
#             raise ValueError('Cannot parse energy unit in Kohn-Sham Gap section.')
#
#     def _parse_homo(self, line):
#         *_,homo_value = line.split()
#         self.homo = mtr.Qty(value=float(homo_value),unit=self.unit)
#
#     def _parse_lumo(self, line):
#         *_,lumo_value = line.split()
#         self.lumo = mtr.Qty(value=float(lumo_value),unit=self.unit)
#
#     def _parse_gap(self, line):
#         *_,gap_value = line.split()
#         self.gap = mtr.Qty(value=float(gap_value),unit=self.unit)
#
# class NetAtomicCharges(Section):
#     def __init__(self):
#         super().__init__(name='NetAtomicCharges')
#         self.unit = None
#         self.charge_dictionary = {}
#
#     def todict(self):
#         return {'charge_dictionary': self.charge_dictionary}
#
#     def parse(self, line):
#         if 'Atom ' in line:
#             self._parse_unit(line=line)
#         elif 'Sum of atomic charges' in line:
#             self._parse_sum_of_atomic_charges(line=line)
#         elif len(line.split()) > 0 and line.split()[0].isdigit():
#             self._parse_atomic_charge(line=line)
#         else:
#             pass
#
#     def _parse_unit(self, line):
#         (charge_unit_string,) = re.search(r'.*\((.*)\).*$',line).groups()
#         if charge_unit_string == 'eV':
#             self.unit = mtr.eV
#         elif charge_unit_string == 'a.u.':
#             self.unit = mtr.au_charge
#         else:
#             raise ValueError('Cannot parse charge unit in Net Atomic Charges section.')
#
#     def _parse_atomic_charge(self, line):
#         atom_number,atomic_species,charge = line.split()
#         self.charge_dictionary[int(atom_number)] = (atomic_species,mtr.Qty(value=float(charge),unit=self.unit))
#
#     def _parse_sum_of_atomic_charges(self, line):
#         *_,sum_of_atomic_charges_str = line.split()
#         self.sum_of_atomic_charges = mtr.Qty(value=float(sum_of_atomic_charges_str),unit=self.unit)
#
# class OrbitalEnergies(Section):
#     def __init__(self):
#         super().__init__(name='OrbitalEnergies')
#         self.alpha = False
#         self.occupied = False
#         self.unit = None
#
#         self.energy_dict = {'alpha': {'occupied': [], 'unoccupied': []},'beta': {'occupied': [], 'unoccupied': []}}
#
#     def todict(self):
#         homo = max(self.energy_dict['alpha']['occupied'],self.energy_dict['beta']['occupied'])
#         lumo = min(self.energy_dict['alpha']['unoccupied'],self.energy_dict['beta']['unoccupied'])
#
#         return {'homo': homo, 'lumo': lumo, **self.energy_dict}
#
#     def parse(self, line):
#         if 'Orbital Energies' in line:
#             self._parse_energy_unit(line=line)
#         elif 'Alpha' in line:
#             self.alpha = True
#         elif 'Beta' in line:
#             self.alpha = False
#         elif 'Occupied' in line:
#             self.occupied = True
#         elif 'Virtual' in line:
#             self.occupied = False
#         else:
#             self._parse_orbital_energies(line=line)
#
#     def _parse_energy_unit(self, line):
#         (unit_str,) = re.search(r'.*\((.*)\).*$',line).groups()
#         if unit_str == 'a.u.':
#             self.unit = mtr.hartree
#         else:
#             raise ValueError('Cannot parse energy unit in Orbital Energies section.')
#
#     def _parse_orbital_energies(self, line):
#         try:
#             energies = [mtr.Qty(value=float(eng),unit=self.unit) for eng in line.strip().split()]
#             self.energy_dict['alpha' if self.alpha else 'beta']['occupied' if self.occupied else 'unoccupied'].extend(energies)
#         except:
#             pass
# #
# # class Polarizability(Section):
# #     def __init__(self):
# #         super().__init__(name='Polarizability')
# #         self.alpha = False
# #         self.occupied = False
# #         self.unit = None
# #
# #         self.energy_dict = {'alpha': {'occupied': [], 'unoccupied': []},'beta': {'occupied': [], 'unoccupied': []}}
# #
# #     def todict(self):
# #         homo = max(self.energy_dict['alpha']['occupied'],self.energy_dict['beta']['occupied'])
# #         lumo = min(self.energy_dict['alpha']['unoccupied'],self.energy_dict['beta']['unoccupied'])
# #
# #         return {'homo': homo, 'lumo': lumo, **self.energy_dict}
# #
# #     def parse(self, line):
# #         if 'Polarizability tensor' in line:
# #             self._parse_polarizability(line=line)
# #         elif 'Alpha' in line:
# #             self.alpha = True
# #         elif 'Beta' in line:
# #             self.alpha = False
# #         elif 'Occupied' in line:
# #             self.occupied = True
# #         elif 'Virtual' in line:
# #             self.occupied = False
# #         else:
# #             self._parse_orbital_energies(line=line)
# #
# #     def _parse_energy_unit(self, line):
# #         (unit_str,) = re.search(r'.*\((.*)\).*$',line).groups()
# #         if unit_str == 'a.u.':
# #             self.unit = mtr.hartree
# #         else:
# #             raise ValueError('Cannot parse energy unit in Orbital Energies section.')
# #
# #     def _parse_orbital_energies(self, line):
# #         try:
# #             energies = [mtr.Qty(value=float(eng),unit=self.unit) for eng in line.strip().split()]
# #             self.energy_dict['alpha' if self.alpha else 'beta']['occupied' if self.occupied else 'unoccupied'].extend(energies)
# #         except:
# #             pass
#
# class SETManTiming(Section):
#     def __init__(self):
#         super().__init__(name='SETManTiming')
#
# # class TDDFTExcitationEnergies(Section):
# #     def __init__(self):
# #         self.section_keywords = {'Excited state': TDDFTExcitation}
#
#     # def _parse(self):
#     #     gen = (l for l in self.lines)
#     #     next(self._yield_excitation_blocks(gen=gen)) # first block is just junk, discard it
#     #
#     #     print(list(self._yield_excitation_blocks(gen=gen)))
#     #     for energy_str,tot_energy_str,mult_str,mom_str,strength_str,*contributions_str in self._yield_excitation_blocks(gen=gen):
#     #          print(energy_str)
#     #          print(energy_str)
#     #
#     # def _yield_excitation_blocks(self, gen):
#     #     #gen = (l for l in self.lines)
#     #     while True:
#     #         t = tuple(itertools.takewhile(lambda l: l.strip() != '',gen))
#     #         if t == ():
#     #             break
#     #         yield t
#
# class SCF(Section):
#     def __init__(self):
#         super().__init__(name='SCF')
#         self.scf_steps = {}
#
#     def todict(self):
#         return {'quadrature_grid': self.quadrature_grid, 'number_threads': self.num_threads,
#                 'cpu_time': self.cpu_time, 'wall_time': self.wall_time, 'scf_energy': self.scf_energy,
#                 'total_energy': self.total_energy, **self.scf_steps}
#
#     def parse(self, line):
#         if 'standard quadrature grid' in line:
#             self._parse_quadrature_grid(line=line)
#         elif 'threads for integral computing' in line:
#             self._parse_num_threads(line=line)
#         elif 'SCF time' in line:
#             self._parse_scf_time(line=line)
#         elif len(line.split()) > 0 and line.split()[0].isdigit():
#             self._parse_scf_step(line=line)
#         elif 'energy in the final basis set' in line:
#             self._parse_energy(line=line)
#         else:
#             pass
#
#     def _parse_energy(self, line):
#         energy_type,energy = re.search(r'\s*(.*?)\s*energy in the final basis set.*?(-?\d*\.\d*)\.*$',line).groups()
#         if energy_type == 'SCF':
#             self.scf_energy = mtr.Qty(value=float(energy),unit=mtr.hartree)
#         elif energy_type == 'Total':
#             self.total_energy = mtr.Qty(value=float(energy),unit=mtr.hartree)
#         else:
#             pass
#
#     def _parse_quadrature_grid(self, line):
#         self.quadrature_grid,*_ = re.search(r'Using (.*) standard quadrature grid.*$',line).groups()
#
#     def _parse_num_threads(self, line):
#         num_threads,*_ = re.search(r'using (\d*) threads for integral computing.*$',line).groups()
#         self.num_threads = int(num_threads)
#
#     def _parse_scf_step(self, line):
#         scf_step_num,energy,error = re.search(r'\s*(\d*)\s*(-?\d*\.\d*)\s*(\d*\.\d*e-?\d*).*$',line).groups()
#         self.scf_steps[int(scf_step_num)] = (mtr.Qty(value=float(energy),unit=mtr.hartree),float(error))
#
#     def _parse_scf_time(self, line):
#         cpu_time,cpu_time_unit,wall_time,wall_time_unit = re.search(r'SCF time:\s*CPU (\d*\.\d*)(.*?)\s*wall (\d*\.\d*)(.*?)\s*$',line).groups()
#         self.cpu_time = mtr.Qty(value=float(cpu_time),unit=getattr(mtr,cpu_time_unit))
#         self.wall_time = mtr.Qty(value=float(wall_time),unit=getattr(mtr,wall_time_unit))
#
# class SCFGradient(Section):
#     def __init__(self):
#         #FIXME: implement parse(self, line), todict(self)
#         super().__init__(name='SCFGradient')
#
# class StandardNuclearOrientation(Section):
#     def __init__(self):
#         super().__init__(name='StandardNuclearOrientation')
#         self.position_unit = None
#         self.atom_dictionary = {}
#
#     def todict(self):
#         return self.atom_dictionary
#
#     def parse(self, line):
#         if 'Standard Nuclear Orientation' in line:
#             self._parse_atomic_position_unit(line=line)
#         elif len(line.split()) > 0 and line.split()[0].isdigit():
#             self._parse_atomic_position(line=line)
#         else:
#             pass
#
#     def _parse_atomic_position_unit(self, line):
#         (unit_str,) = re.search(r'.*\((.*)\).*$',line).groups()
#         if unit_str == 'Angstroms':
#             self.position_unit = mtr.angstrom
#         else:
#             raise ValueError('Cannot parse atomic position unit in Standard Nuclear Orientation section.')
#
#     def _parse_atomic_position(self, line):
#         try:
#             atom_number,atomic_symbol,x,y,z = line.strip().split()
#             atomic_position_value = np.array([float(x),float(y),float(z)]).reshape(3,1)
#             self.atom_dictionary[int(atom_number)] = (atomic_symbol,mtr.Qty(value=atomic_position_value,unit=self.position_unit))
#         except ValueError: # wrong line got passed into this function, so just pass without processing
#             pass
#
# class TDDFTExcitation(Section):
#     def __init__(self):
#         super().__init__(name='TDDFTExcitation')
#
#     def parse(self, line):
#         if 'Excited state' in line:
#             self._parse_excited_state(line=line)
#         elif 'Total energy for state' in line:
#             self._parse_excited_total_energy(line=line)
#         elif 'Multiplicity:' in line:
#             self._parse_multiplicity(line=line)
#         elif 'Trans. Mom.' in line:
#             self._parse_transition_dipole(line=line)
#         elif 'Strength' in line:
#             self._parse_oscillator_strength(line=line)
#
#     def todict(self):
#         return {'excitation_energy': self.excitation_energy, 'total_energy': self.total_energy,
#                 'multiplicity': self.multiplicity, 'transition_dipole': self.transition_dipole,
#                 'oscillator_strength': self.oscillator_strength}
#
#     def _parse_excited_state(self, line):
#         (energy_unit_str,energy_value_str) = re.search(r'.*\((.*)\).*=\s*(\d*).*$',line).groups()
#         if energy_unit_str == 'eV':
#             excitation_energy_unit = mtr.eV
#         else:
#             raise ValueError('Cannot parse excitation energy unit in TDDFT Excitation section.')
#         energy_value = float(energy_value_str)#float(line.strip().split()[-1])
#         self.excitation_energy = mtr.Qty(value=energy_value,unit=excitation_energy_unit)
#
#     def _parse_excited_total_energy(self, line):
#         *_,tot_energy_val_str,tot_energy_unit_str = line.strip().split()
#         if tot_energy_unit_str == 'au':
#             tot_energy_unit = mtr.hartree
#         else:
#             raise ValueError('Cannot parse total energy unit in TDDFT Excitation section.')
#         self.total_energy = mtr.Qty(value=float(tot_energy_val_str),unit=tot_energy_unit)
#
#     def _parse_multiplicity(self, line):
#         *_,self.multiplicity = line.lower().strip().split()
#
#     def _parse_transition_dipole(self, line):
#         *_,x,_,y,_,z,_ = line.strip().split()
#         self.transition_dipole = mtr.Qty(value=np.array([float(x),float(y),float(z)]),unit=mtr.debye) # Unit is debye as stated here: http://iopenshell.usc.edu/forum/topic.php?id=3696
#
#     def _parse_oscillator_strength(self, line):
#         *_,strength_val_str = line.strip().split()
#         self.oscillator_strength = float(strength_val_str)
#
#     def _parse(self):
#         #FIXME: fix all this
#         energy_str,tot_energy_str,mult_str,mom_str,strength_str,*contributions_str = self.lines
#
#         # EXCITATION ENERGY
#
#
#
#         # EXCITED TOTAL ENERGY
#
#
#
#         # MULTIPLICITY
#
#
#
#         # TRANSITION DIPOLE
#
#
#
#         # OSCILLATOR STRENGTH
#
#
#         # OCCUPIED-UNOCCUPIED ORBITAL PAIR CONTRIBUTIONS
#
#         contributions_dict = {}
#         for l in contributions_str:
#             stripped = l.strip()
#             if stripped != '':
#                 try:
#                     f,t = re.search(r'.*\((.*)\).*\((.*)\).*$',l).groups()
#                     *_,amplitude_str = stripped.split()
#                     contributions_dict[int(f),int(t)] = float(amplitude_str)
#                 except AttributeError:
#                     break
#
#         return excitation_energy,total_energy,multiplicity,transition_moment,strength,contributions_dict
#
# class TDDFTTDARoots(Section):
#     def __init__(self):
#         super().__init__(name='TDDFTTDARoots')
#
#     def parse(self, line):
#         pass



from __future__ import annotations
from typing import Dict, IO, Optional, Tuple, Union

import contextlib
import itertools
import numpy as np
import materia as mtr
from materia.utils import memoize
import networkx as nx
import openbabel as ob
import pubchempy as pcp
import rdkit, rdkit.Chem, rdkit.Chem.AllChem
import scipy.linalg
import tempfile

__all__ = ["Structure"]


class Structure:
    def __init__(self, *atoms: mtr.Atom) -> None:
        self.atoms = tuple(atoms)

    @staticmethod
    def read(filepath: str) -> Structure:
        """
        Read structure from a file.

        Args:
            filepath: Path to file from which the structure will be read. Can be an absolute or a relative path.

        Returns:
            materia.Structure: Structure object based on file contents.

        Raises:
            ValueError: If structure file extension is not recognized.
        """
        if filepath.endswith(".xyz"):
            return _read_xyz(filepath=filepath)
        else:
            raise ValueError("Cannot read file with given extension.")

    def __add__(self, other: mtr.Structure) -> mtr.Structure:
        return mtr.Structure((*self.atoms, *other.atoms))

    @property
    def bonds(self) -> Dict[int, int]:
        obmol = ob.OBMol()

        bonds = {k: [] for k in range(self.num_atoms)}

        for a in self.atoms:
            obatom = ob.OBAtom()
            obatom.SetAtomicNum(a.Z)
            obatom.SetVector(*a.position.squeeze())
            obmol.AddAtom(obatom)

        obmol.ConnectTheDots()
        obmol.PerceiveBondOrders()

        for bond in ob.OBMolBondIter(obmol):
            a, b = bond.GetBeginAtomIdx() - 1, bond.GetEndAtomIdx() - 1
            bonds[a].append(b)
            bonds[b].append(a)

        return bonds

    def to_graph(self, explicit_hydrogen: Optional[bool] = True) -> nx.Graph:
        g = nx.Graph()

        bonds = self.bonds

        if not explicit_hydrogen:
            non_hydrogens = [i for i, Z in enumerate(self.atomic_numbers) if Z != 1]
            g.add_nodes_from(non_hydrogens)

            edges = [
                (i, j)
                for i, v in bonds.items()
                for j in v
                if (i in non_hydrogens and j in non_hydrogens)
            ]
            g.add_edges_from(edges)

            hydrogens = [i for i, Z in enumerate(self.atomic_numbers) if Z == 1]

            nx.set_node_attributes(
                G=g,
                values={
                    i: {
                        "Z": self.atomic_numbers[i],
                        "position": self.atomic_positions[:, i],
                        "num_H": sum(1 for x in bonds[i] if x in hydrogens),
                    }
                    for i in non_hydrogens
                },
            )

            g = nx.relabel_nodes(
                g, {k: v for k, v in zip(g.nodes, range(len(g.nodes)))}
            )
        else:
            g.add_nodes_from(range(len(self.atomic_numbers)))

            edges = [(i, j) for i, v in bonds.items() for j in v]
            g.add_edges_from(edges)

            nx.set_node_attributes(
                G=g,
                values={
                    i: {"Z": Z, "position": self.atomic_positions[:, i]}
                    for i, Z in enumerate(self.atomic_numbers)
                },
            )

        return g

    @staticmethod
    def retrieve(
        name: Optional[str] = None,
        smiles: Optional[str] = None,
        inchi: Optional[str] = None,
        inchikey: Optional[str] = None,
    ) -> mtr.Structure:
        kwargs = (
            (name, "name"),
            (smiles, "smiles"),
            (inchi, "inchi"),
            (inchikey, "inchikey"),
        )
        try:
            identifier, identifier_type = next(
                (k, v) for k, v in kwargs if k is not None
            )
        except StopIteration:
            raise ValueError(
                "Identifier (name, SMILES, InChi, or InChiKey) must be provided to retrieve structure."
            )
        try:
            # this just picks the first returned compound; if there are multiple, we are assuming that the first such compound is the "most relevant" in some sense
            cid, *_ = pcp.get_cids(identifier, identifier_type)
            if cid == 0:
                raise ValueError
        except (ValueError, OSError):
            raise ValueError(f"Structure retrieval for {identifier} failed.")

        try:
            return _structure_from_pubchem_compound(
                compound=pcp.Compound.from_cid(cid, record_type="3d")
            )
        except pcp.NotFoundError:
            # no 3d structure from pubchem; there must be a 2d structure since a cid was found
            [property_dict] = pcp.get_properties(
                properties="IsomericSMILES", identifier=cid, namespace="cid"
            )
            return Structure.generate(smiles=property_dict["IsomericSMILES"])

    @staticmethod
    def generate(
        name: Optional[str] = None,
        smiles: Optional[str] = None,
        inchi: Optional[str] = None,
        inchikey: Optional[str] = None,
    ) -> mtr.Structure:
        kwargs = (
            (name, "name"),
            (smiles, "smiles"),
            (inchi, "inchi"),
            (inchikey, "inchikey"),
        )
        try:
            identifier, identifier_type = next(
                (k, v) for k, v in kwargs if k is not None
            )
        except StopIteration:
            raise ValueError(
                "Identifier (name, SMILES, InChi, or InChiKey) must be provided to generate structure."
            )

        if identifier_type == "smiles":
            return _structure_from_identifier(smiles=smiles)
        elif identifier_type == "inchi":
            return _structure_from_identifier(inchi=inchi)
        else:
            raise ValueError(f"Structure generation for {identifier} failed.")

    def write(self, file: Union[str, IO], overwrite: Optional[bool] = False) -> None:
        """
        Write structure to a file.

        Args:
            file: Path to file to which the structure will be written. Can be an absolute or a relative path.
            
            overwrite: If False, an error is raised if `filepath` already exists and the structure is not written. Ignored if `file` is a file-like object. Defaults to False.
        """
        open_code = "w" if overwrite else "x"
        with open(mtr.expand(file), open_code) if isinstance(
            file, str
        ) else contextlib.nullcontext(file) as f:
            if f.name.endswith(".xyz"):
                s = self.to_xyz()
            else:
                raise ValueError("Cannot write to file with given extension.")

            try:
                f.write(s)
            except TypeError:
                f.write(s.encode())

            f.flush()

    @contextlib.contextmanager
    def tempfile(self, suffix: str, dir: Optional[str] = None):
        with tempfile.NamedTemporaryFile(
            dir=mtr.expand(dir) if dir is not None else None, suffix=suffix
        ) as fp:
            try:
                self.write(file=fp)
                yield fp
            finally:
                pass

    def to_xyz(self) -> str:
        return f"{self.num_atoms}\n\n" + "\n".join(
            f"{atom} {x} {y} {z}"
            for atom, (x, y, z) in zip(
                self.atomic_symbols, self.atomic_positions.T.value
            )
        )

    def to_rdkit(self, charge: Optional[int] = 0) -> rdkit.Chem.rdchem.Mol:
        return mtr.xyz2mol(self.atomic_numbers, charge, self.atomic_positions.T)

    @property
    def num_atoms(self) -> int:
        return len(self.atoms)

    @property
    @memoize
    def atomic_symbols(self) -> Tuple[str]:
        return tuple(atom.atomic_symbol for atom in self.atoms)

    @property
    @memoize
    def atomic_positions(self) -> mtr.Qty:
        value = np.hstack([atom.position.value for atom in self.atoms])
        unit_set = set(atom.position.unit for atom in self.atoms)
        try:
            (unit,) = tuple(unit_set)
        except ValueError:
            raise ValueError("Atomic positions do not have a common unit.")

        return value * unit

    @property
    @memoize
    def atomic_numbers(self) -> Tuple[int]:
        return tuple(atom.Z for atom in self.atoms)

    @property
    @memoize
    def atomic_masses(self) -> mtr.Qty:
        value = tuple(atom.mass.value for atom in self.atoms)
        unit_set = set(atom.mass.unit for atom in self.atoms)
        try:
            (unit,) = tuple(unit_set)
        except ValueError:
            raise ValueError("Atomic masses do not have a common unit.")

        return value * unit

    @property
    @memoize
    def mass(self) -> mtr.Qty:
        value = sum(self.atomic_masses.value)
        unit = self.atomic_masses.unit

        return value * unit

    @property
    @memoize
    def center_of_mass(self) -> mtr.Qty:
        return (
            (self.atomic_masses.value * self.atomic_positions.value)
            .sum(1)
            .reshape(3, 1)
            * self.atomic_positions.unit
            / self.mass.value
        )

    @property
    @memoize
    def centered_atomic_positions(self) -> mtr.Qty:
        return self.atomic_positions - self.center_of_mass

    @property
    @memoize
    def inertia_tensor(self) -> mtr.Qty:
        atomic_masses = self.atomic_masses
        centered_atomic_positions = self.centered_atomic_positions

        xx, xy, xz, yy, yz, zz = (
            np.dot(atomic_masses.value, a * b)
            for a, b in itertools.combinations_with_replacement(
                centered_atomic_positions.value, r=2
            )
        )

        I = np.zeros((3, 3))
        I[np.tril_indices(3, -1)] = -xy, -xz, -yz
        I += I.T + np.diag([yy + zz, xx + zz, xx + yy])

        unit = atomic_masses.unit * centered_atomic_positions.unit ** 2

        return I * unit

    @property
    @memoize
    def inertia_aligned_atomic_positions(self) -> mtr.Qty:
        # FIXME: examine and clean this one up
        if self.num_atoms == 1:
            # i.e. this is an atomic species
            return np.eye(3)

        principal_moments, principal_directions = scipy.linalg.eigh(
            self.inertia_tensor.value
        )

        sorted_moments, sorted_directions = zip(
            *sorted(zip(principal_moments, principal_directions.T), reverse=True)
        )
        u, v, _ = sorted_directions

        u /= scipy.linalg.norm(u)
        v /= scipy.linalg.norm(v)
        z = np.array([[0, 0, 1]]).T
        y = np.array([[0, 1, 0]]).T
        x = np.array([[1, 0, 0]]).T
        axis = np.cross(u.T, z.T).T
        c = np.dot(u.T, z)  # cosine of the angle between u and z
        u1, u2, u3 = np.ravel(axis)
        K = np.array([[0, -u3, u2], [u3, 0, -u1], [-u2, u1, 0]])
        s = np.sqrt(1 - c ** 2)
        Ru = (np.eye(3) + s * K + (1 - c) * (K @ K)).astype("float64")
        axis = np.cross((Ru @ v).T, y.T).T
        axis /= scipy.linalg.norm(axis)
        c = np.dot((Ru @ v).T, y)  # cosine of the angle between Ru@v and y
        u1, u2, u3 = np.ravel(axis)
        K = np.array([[0, -u3, u2], [u3, 0, -u1], [-u2, u1, 0]])
        s = np.sqrt(1 - c ** 2)
        Rv = (np.eye(3) + s * K + (1 - c) * (K @ K)).astype("float64")
        R = Rv @ Ru

        print([[x for x in v] for v in R])

        Rp = np.hstack([y, z, x]) @ np.linalg.inv(
            np.hstack(
                [u.reshape(3, 1), v.reshape(3, 1), np.cross(u.T, v.T).reshape(3, 1)]
            )
        )

        print([[x for x in v] for v in Rp])
        print(R @ u)
        print(R @ v)
        print(Rp @ u)
        print(Rp @ v)
        print(R @ R.T)
        print(Rp @ Rp.T)

        return (
            R @ self.centered_atomic_positions.value
        ) * self.centered_atomic_positions.unit

    @property
    @memoize
    def principal_moments(self) -> mtr.Qty:
        scipy.linalg.eigvalsh(self.inertia_tensor) * self.inertia_tensor.unit

    @property
    @memoize
    def principal_axes(self) -> Tuple[float]:
        _, axes = scipy.linalg.eigh(self.inertia_tensor)
        return tuple(mtr.normalize(ax) for ax in axes.T)

    @property
    @memoize
    def is_linear(self) -> bool:
        (m1, m2, m3) = self.principal_moments.value / sum(self.principal_moments.value)
        return (
            (m1 == 0 and m2 == m3) or (m2 == 0 and m1 == m3) or (m3 == 0 and m1 == m2)
        )

    @property
    @memoize
    def is_planar(self) -> mtr.Qty:
        (m1, m2, m3) = self.principal_moments.value / sum(self.principal_moments.value)
        return (m1 + m2 == m3) or (m1 + m2 == m3) or (m1 + m2 == m3)

    @property
    @memoize
    def diameter(self) -> mtr.Qty:
        hull = scipy.spatial.ConvexHull(self.atomic_positions.value)
        # only look at atoms on the convex hull
        kdt = scipy.spatial.KDTree(self.atomic_positions.value[hull.vertices, :])

        # return maximum pairwise distance among all atoms on the convex hull
        return (
            max(kdt.sparse_distance_matrix(kdt, np.inf).values())
            * self.atomic_positions.unit
        )

    # FIXME: fix this, annotation too
    @property
    @memoize
    def pointgroup(self):
        sf = mtr.symfinder.SymmetryFinder()
        return sf.molecular_pointgroup(
            atomic_positions=self.atomic_positions.value,
            atomic_numbers=self.atomic_numbers,
        )

    # FIXME: fix this, annotation too
    @property
    @memoize
    def maximally_symmetric_spanning_set(self):
        """
        Finds a set of vectors which span R^3 and which are related to one another
        as much as possible by symmetry operations of the molecule whose
        atomic species and atomic positions are given by xyz.

        Parameters
        ----------
        xyz : XYZ
            XYZ object containing the atomic species and atomic_positions of the molecule
            whose maximally symmetric spanning set is to be computed.

        Returns
        -------
        dict
            Dictionary containing three entries: axes, whose value is a list
            containing the vectors in the maximally symmetric spanning set;
            number_of_equivalent_axes, whose value is the number of axes in the
            spanning set which are related to one another by a symmetry rotation;
            and wprime.
        """
        axgen = mtr.symfinder.AxesGenerator()

        return axgen.generate_axes(
            pointgroup_symbol=self.pointgroup, inertia_tensor=self.inertia_tensor
        )


# ----- IO helper functions ----- #


def _read_xyz(filepath: str, coordinate_unit: str = "angstrom") -> Structure:
    with open(mtr.expand(filepath), "r") as f:
        atom_data = np.atleast_2d(
            np.loadtxt(
                fname=f,
                usecols=(0, 1, 2, 3),
                skiprows=1,
                max_rows=int(next(f)),
                dtype=str,
            )
        )

    atomic_symbols = atom_data[:, 0]
    atomic_positions = (
        np.asarray(p, dtype="float64") * getattr(mtr, coordinate_unit)
        for p in atom_data[:, 1:]
    )
    atoms = (
        mtr.Atom(element=symbol, position=position)
        for symbol, position in zip(atomic_symbols, atomic_positions)
    )

    return Structure(*atoms)


def _structure_from_pubchem_compound(compound: pcp.Compound) -> mtr.Structure:
    # FIXME: assumes the pubchem distance unit is angstrom - is this correct??
    atom_generator = (
        (a.element, (a.x, a.y, a.z) * mtr.angstrom) for a in compound.atoms
    )
    atoms = (mtr.Atom(element=symb, position=pos) for symb, pos in atom_generator)

    return mtr.Structure(*atoms)


def _structure_from_identifier(
    smiles: Optional[str] = None, inchi: Optional[str] = None, num_conformers: int = 25
) -> mtr.Structure:
    # for motivation on generating 25 (as opposed to, say, 10 or 100) conformers, see:
    # https://github.com/rdkit/UGM_2015/blob/master/Presentations/ETKDG.SereinaRiniker.pdf

    if smiles is not None:
        mol = rdkit.Chem.MolFromSmiles(smiles, sanitize=False)
    elif inchi is not None:
        mol = rdkit.Chem.MolFromInchi(inchi, sanitize=False)
    else:
        raise ValueError("Either SMILES or InChi required to generate structure.")

    # sanitize
    try:
        mol.UpdatePropertyCache(False)
        mol = rdkit.Chem.Mol(mol.ToBinary())
        rdkit.Chem.SanitizeMol(mol)
    except ValueError:
        raise ValueError("Cannot sanitize RDKit molecule.")

    # hydrogenate
    mol = rdkit.Chem.AddHs(mol)

    # embed to generate 3D coords
    embedding_parameters = rdkit.Chem.AllChem.ETKDG()
    embed_return_code = rdkit.Chem.AllChem.EmbedMolecule(
        mol=mol, params=embedding_parameters
    )

    if embed_return_code == -1:
        embedding_parameters.useRandomCoords = True
        rdkit.Chem.AllChem.EmbedMolecule(mol=mol, params=embedding_parameters)

    # embed multiple conformers and find one with lowest energy
    rdkit.Chem.AllChem.EmbedMultipleConfs(
        mol, numConfs=num_conformers, params=embedding_parameters
    )

    # MMFF seems to give slightly better geometries, so it is preferred if possible
    if rdkit.Chem.AllChem.MMFFHasAllMoleculeParams(mol=mol):
        mmff_props = rdkit.Chem.AllChem.MMFFGetMoleculeProperties(mol=mol)
        rdkit.Chem.AllChem.MMFFSanitizeMolecule(mol=mol)
        energy = lambda conformer: rdkit.Chem.AllChem.MMFFGetMoleculeForceField(
            mol=conformer.GetOwningMol(),
            pyMMFFMolProperties=mmff_props,
            confId=conformer.GetId(),
        ).CalcEnergy()
    else:
        energy = lambda conformer: rdkit.Chem.AllChem.UFFGetMoleculeForceField(
            mol=conformer.GetOwningMol(), confId=conformer.GetId()
        ).CalcEnergy()

    energies = {conformer: energy(conformer) for conformer in mol.GetConformers()}

    conformer = min(energies, key=energies.get)

    # convert to Structure
    symbols = (a.GetSymbol() for a in conformer.GetOwningMol().GetAtoms())

    # FIXME: assumes the RDKIT distance unit is angstrom - is this correct??
    # NOTE: using conformer.GetPositions sometimes causes a seg fault (RDKit) - use GetAtomPosition instead
    atoms = (
        mtr.Atom(element=symbol, position=conformer.GetAtomPosition(i) * mtr.angstrom,)
        for i, symbol in enumerate(symbols)
    )

    return mtr.Structure(*atoms)



from __future__ import annotations
import inspect
import materia
from typing import Any, Iterable, Union

__all__ = ["CCDCOutput"]


class CCDCOutput:
    """
    Interface for outputs from tasks run with CCDC.

    Attributes:
        filepath (str): Absolute path to file from which output will be read.
    """

    def __init__(self, filepath: str) -> None:
        """
        Args:
            filepath: Path to file from which output will be read. Can be an absolute or a relative path.
        """
        self.filepath = materia.expand(filepath)

        raise NotImplementedError

    def get(self, *args: str) -> Union[Any, Iterable[Any]]:
        """
        Get quantities from parsed CCDC output.

        Args:
            *args: Quantities to get from parsed output.
        """
        method_dict = dict(inspect.getmembers(self, predicate=inspect.isroutine))

        with open(self.filepath, "r") as f:
            lines = "".join(f.readlines())

        quantities = tuple(method_dict[name.lower()](lines=lines) for name in args)

        if len(quantities) == 1:
            return quantities[0]
        return quantities
